#include "binary.h"

using namespace arc;

static const QString		ARC_NAME("Binary");
static const int			LEFT_ID = 0;
static const int			RIGHT_ID = 1;

static void mult(float* dst, const float* src, const int size);

/* BINARY
 ******************************************************************/
Binary::Binary(const Binary& o)
	: Arc(o)
	, mOp(o.mOp)
	, mLeft(o.mLeft)
	, mRight(o.mRight)
{
	mLeft.setId(LEFT_ID);
	mRight.setId(RIGHT_ID);
}

Binary::Binary(const int op)
	: mOp(op)
{
	mLeft.setId(LEFT_ID);
	mRight.setId(RIGHT_ID);
}

Binary::Binary(const int op, Arc* left, Arc* right)
	: mOp(op)
{
	mLeft.setId(LEFT_ID);
	mRight.setId(RIGHT_ID);

	mLeft.setArc(left);
	mRight.setArc(right);
}

Binary* Binary::copy() const
{
	return new Binary(*this);
}

void Binary::getParams(param::List& list, param::GetArgs& get) const
{
	mLeft.getParams(list, get);
	mRight.getParams(list, get);
}

param::Parent* Binary::findParent(param::UrlIterator& it)
{
	if (it.matches(LEFT_ID)) {
		++it;
		return mLeft.findParent(it);
	} else if (it.matches(RIGHT_ID)) {
		++it;
		return mRight.findParent(it);
	}
	return NULL;
}

void Binary::setParam(const param::UrlIterator&, param::SetArgs&)
{
}

void Binary::work(WorkParams& p, float* row)
{
	if (mLeft.isEmpty() || mRight.isEmpty()) return;

	// I have two sides that are evalulated independently,
	// so I need to operate on a spare buffer.
	Mask*			bm = p.acquireScratch();
	if (bm == NULL) return;

	float*			b = bm->get(p.mRowSize);
	if (b != NULL) {
		// Run my sides
		memcpy(b, row, p.mRowSize*sizeof(float));
		mLeft.getArc()->work(p, row);
		mRight.getArc()->work(p, b);
		// Apply my op
		if (mOp == OP_MULT) mult(row, b, p.mRowSize);
	}

	p.releaseScratch(bm);
}

int Binary::forEach(arc::Visitor& v)
{
	Arc::forEach(v);
	mLeft.forEach(v);
	mRight.forEach(v);
	return CONTINUE;
}

#ifdef QT_DEBUG
void Binary::printTree(const int tabs) const
{
	{
		QDebug		d = qDebug();
		for (int k=0; k<tabs; k++) (d.nospace()) << "\t";
		(d.nospace()) << ARC_NAME;
	}
	{
		QDebug		d = qDebug();
		for (int k=0; k<tabs; k++) (d.nospace()) << "\t";
		(d.nospace()) << "  left id=";
		(d.nospace()) << mLeft.getId();
	}
	mLeft.printTree(tabs+1);
	{
		QDebug		d = qDebug();
		for (int k=0; k<tabs; k++) (d.nospace()) << "\t";
		(d.nospace()) << "  right id=";
		(d.nospace()) << mRight.getId();
	}
	mRight.printTree(tabs+1);
}
#endif

/* Misc
 ******************************************************************/
static void mult(float* dst, const float* src, const int size)
{
	const float*		end = dst+size;
	while (dst != end) {
		(*dst) *= (*src);
		dst++;
		src++;
	}
}
