#include "sine.h"

#include <assert.h>
#include <math.h>

using namespace arc;

static const QString		ARC_NAME("Sine");
static const QString		PERIOD_NAME("Period");
static const int			PERIOD_ID = 0;
static const QString		PHASE_NAME("Phase");
static const int			PHASE_ID = 1;
//static const QString		POLARITY_NAME("Polarity");
//static const int			POLARITY_ID = 2;

// Harcoded limit on period, because somehow the range has to be
// converted to a scaled 0-1 to make it easy for clients.
static const float			PERIOD_MAX = 10;

static const float			PI = 3.1415926535f;

static inline float fromRaw(const float raw) {
	return PI * 2 * raw;
}

/* SINE
 ******************************************************************/
Sine::Sine(const float period, const float phase, const bool bipolar)
	: mPeriod(period)
	, mPhase(phase)
	, mBipolar(bipolar)
{
}

Sine::Sine(const Sine& o)
	: Arc(o)
	, mPeriod(o.mPeriod)
	, mPhase(o.mPhase)
	, mBipolar(o.mBipolar)
{
}

Sine* Sine::copy() const
{
	return new Sine(*this);
}

void Sine::getParams(param::List& list, param::GetArgs& get) const
{
	list.addParam(&ARC_NAME, PERIOD_NAME, PERIOD_ID, mPeriod / PERIOD_MAX, get);
	list.addParam(&ARC_NAME, PHASE_NAME, PHASE_ID, mPhase, get);
}

param::Parent* Sine::findParent(param::UrlIterator& it)
{
	assert(it.hasValue());
	if (!it.hasValue()) return NULL;
	return this;
}

void Sine::setParam(const param::UrlIterator& it, param::SetArgs& sa)
{
	if (it.matches(PERIOD_ID)) {
		float		trans = sa.mV * PERIOD_MAX;
		if (trans < 0) trans = 0;
		else if (trans > PERIOD_MAX) trans = PERIOD_MAX;
		if (mPeriod == trans) return;		// no change
		mPeriod = trans;
	} else if (it.matches(PHASE_ID)) {
		if (mPhase == sa.mV) return;	// no change
		mPhase = sa.mV;
	} else {
		return;
	}
	sa.mChanged = true;
}

void Sine::work(WorkParams& p, float* row)
{
	float		period = mPeriod, phase = mPhase;
	ModEval*	mod = p.acquireEval(this, 0, &period, &phase);

	float				v;
	for (int k=0; k<p.mRowSize; k++) {
		if (mod != NULL) mod->moveTo(k);

		if (row[k] != 0) {
			v = sinf(fromRaw((row[k]*period) + phase));
			if (!mBipolar && v < 0) v *= -1;
			row[k] = v;
		}
	}

	if (mod != NULL) p.releaseEval(mod);
}

#ifdef QT_DEBUG
void Sine::printTree(const int tabs) const
{
	QDebug		d = qDebug();
	for (int k=0; k<tabs; k++) (d.nospace()) << "\t";

	(d.nospace()) << ARC_NAME;
	(d.nospace()) << " id=";
	(d.nospace()) << mId;
	(d.nospace()) << " period=";
	(d.nospace()) << mPeriod;
	(d.nospace()) << " phase=";
	(d.nospace()) << mPhase;
}
#endif
