#ifndef ARCDEFS_H
#define ARCDEFS_H

#include <stdlib.h>

#include "collections/recyclelist.h"
#include "engine/archead.h"
#include "math/primitives/range.h"
#include "compositing/mask.h"

namespace arc {

// Types of sources
static const int		DISTANCE_SRC = 0;
static const int		NORMAL_SRC = 1;
static const int		SRC_SIZE = NORMAL_SRC+1;

static const int		HAS_DISTANCE_SRC = (1<<0);
static const int		HAS_NORMAL_SRC = (1<<1);
static const int		HAS_ALL_SRC = (HAS_DISTANCE_SRC | HAS_NORMAL_SRC);

// For visting
static const int		CONTINUE = 1;
static const int		STOP = 0;

static const int		MAX_EVAL = 4;

// Forward declarations
class ModEval;

/* VISITOR
 * For walking the Arc graph.
 ******************************************************************/
class Visitor {
public:
	virtual ~Visitor()	{ }

	// Answer CONTINUE, etc.
	virtual int		visit(Arc&) = 0;

protected:
	Visitor()		{ }
};

/* WORK-PARAMS
 * Store the modulations.
 ******************************************************************/
class WorkParams {
public:
	// Only valid after calling begin()
	float*					mDistanceRow;
	float*					mNormalRow;
	int						mRowSize;

public:
	WorkParams();
	~WorkParams();

	// This is a hack to make sure I'm set up and ready for
	// use on the given arc and size.  If successful, then,
	// I will have mDistanceRow and mNormalRow values at the
	// correct size.  Flags needs to be set to the types of
	// sources the client will supply (HAS_DISTANCE_SRC etc).
	bool					begin(	ArcHead*, const int rowSize,
									const int flags);

	// Use i.e. a SRC flag, like HAS_DISTANCE_SRC.
	bool					hasFlag(const int flag) const;

	// Currently clients can have up to 4 evaluated
	// parameters.  It would be nice if this was implemented
	// as a variadic so clients could have any number, but this
	// gets called so often I worry about the performance.  The
	// work around would be to just grab more eval blocks, with
	// successive blocks using a non-0 idx to indicate the index
	// of v0.
	ModEval*				acquireEval(Arc* arc, const int idx,
										float* v0, float* v1 = NULL,
										float* v2 = NULL, float* v3 = NULL);
	void					releaseEval(ModEval*);

	// Clients can grab buffers of data for temporary work.
	Mask*					acquireScratch();
	void					releaseScratch(Mask*);

private:
	ArcHead*				mArc;
	int						mFlags;
	// The various sources
	ModSrc					mSrc[SRC_SIZE];
	// Cache evals that aren't in use.
	RecycleList<ModEval*>	mEval;
	// Store buffers of data for clients to use.
	RecycleList<Mask*>		mScratch;


	// A temporary list to turn the arguments into an array.  The
	// clients can't have an array, it doesn't really make sense,
	// but it's useful when building the eval.
	float*					mDst[MAX_EVAL];

	void setTargets(	ModEval* eval, const int arcId,
						const int startIdx, float** v);
};

/* MOD-EVAL
 * Store all data necessary for a source to modulate a
 * destination.  A destination is a single *float64.  A
 * source is a slice of precomputed float64 values.
 ******************************************************************/
class ModEval {
public:
	ModEval();

	void				moveTo(const int rowIdx);

protected:
	friend class WorkParams;
	int					mSize;

private:
	ModEval(const ModEval&);
	void				operator=(const ModEval&);

	int					mFlags[MAX_EVAL];
	float*				mSrcRow[MAX_EVAL];
	float*				mDst[MAX_EVAL];
	float				mOrig[MAX_EVAL];		// The initialized values for each destination
	Range				mR[MAX_EVAL];
};

} // arc

#endif // ARCDEFS_H
