#include "fxdefs.h"

#include "fx.h"
#include "math/primitives/segment3.h"

using namespace fx;

static float poly_dist(const Point3& pt, const QPolygonF& poly);

/* FX-DEFS::VISIT-PARAMS
 ******************************************************************/
VisitParams::VisitParams()
	: mDepth(0)
{
}

/* WORK-ARGS.  Arguments to work().
 ******************************************************************/
WorkArgs::WorkArgs()
	: mSize(0, 0)
{
	uncacheAll();
}

WorkArgs::~WorkArgs()
{
}

bool WorkArgs::begin(const QSize& s)
{
	const int			cellSize = s.width() * s.height();
	const int			imgSize = cellSize*4;
	if (mSize != s) {
		if (mCell.get(cellSize) == NULL) return false;
		if (mImg.get(imgSize) == NULL) return false;
		mSize = s;
		// Only uncache by default when the size changes, or
		// else the pointing of caching is lost.
		uncacheAll();
	}

	float*				img = mImg.get(imgSize);
	if (img == NULL) return false;

	set_to_0(img, imgSize);
	return true;
}

Cell* WorkArgs::getCell()
{
	return mCell.get(mSize.width() * mSize.height());
}

float* WorkArgs::getImg()
{
	return mImg.get(mSize.width() * mSize.height() * 4);
}

void WorkArgs::setCached(const int flag, const bool on)
{
	if (on) mCached |= flag;
	else mCached &= ~flag;
}

void WorkArgs::uncacheAll()
{
	mCached = 0;
	mMinD = 0;
	mMaxD = 0;
}

void WorkArgs::cacheDistance(float* min, float* max)
{
	if ((mCached&DISTANCE_CACHED) != 0) {
		if (min != NULL) *min = mMinD;
		if (max != NULL) *max = mMaxD;
		return;
	}

	const int			w = mSize.width(), h = mSize.height();
	Cell*				cell = mCell.get(w*h);
	if (cell == NULL) return;
	// XXX Should clear it out, or do distance to point
	if (mShape.size() < 2) return;

	int					pix = 0;
	Point3				pt(0, 0, 0);
	QPointF				qpt;
	mMinD = 1000000000;
	mMaxD = -1000000000;
	for (int y=0; y<h; y++) {
		pt.y = y;
		qpt.setY(pt.y);
		for (int x=0; x<w; x++) {
			pt.x = x;
			qpt.setX(pt.x);
			float		d = poly_dist(pt, mShape);
			if (!mShape.containsPoint(qpt, Qt::OddEvenFill)) d = -d;
			if (d < mMinD) mMinD = d;
			if (d > mMaxD) mMaxD = d;
			cell[pix].mD = d;
			pix++;
		}
	}
	setCached(DISTANCE_CACHED, true);
	if (min != NULL) *min = mMinD;
	if (max != NULL) *max = mMaxD;
}

// Convert an alpha premultiplied colour to the local colorspace,
// a packed 32 bit uint.
static inline uint unmatte(const float rm, const float gm, const float bm, const float am)
{
	uint		r = 0, g = 0, b = 0, a = (uint)(am*255.0f);
	if (am <= 0) {
		r = (uint)(rm*255.0f);
		g = (uint)(gm*255.0f);
		b = (uint)(bm*255.0f);
	} else {
		r = (uint)((rm*255.0f) / am);
		g = (uint)((gm*255.0f) / am);
		b = (uint)((bm*255.0f) / am);
	}
	// Limit the components
	if (r > 255) r = 255;
	if (g > 255) g = 255;
	if (b > 255) b = 255;
	if (a > 255) a = 255;
	// Pack
	return (a << 24) | (r << 16) | (g << 8) | b;
}

void WorkArgs::finish(QImage& img)
{
	const int				w = mSize.width(), h = mSize.height();
	int						spix = 0;
	const float*			src = mImg.get(w*h*4);
	if (src == NULL) return;
	for (int y=0; y<h; y++) {
		for (int x=0; x<w; x++) {
			img.setPixel(x, y, unmatte(src[spix], src[spix+1], src[spix+2], src[spix+3]));
			spix += 4;
		}
	}
}

Mask* WorkArgs::acquireMask(const int size)
{
	Mask*		ans = NULL;
	if (!mScratch.isEmpty()) ans = mScratch.popFront();
	if (ans == NULL) ans = new Mask();
	if (ans != NULL && ans->get(size) != NULL) return ans;
	delete ans;
	return NULL;
}

void WorkArgs::releaseMask(Mask* m)
{
	if (m == NULL) return;
	mScratch.pushFront(m);
}

/* WORK-OUTPUT.  Store the result of work processing.
 ******************************************************************/
WorkOutput::WorkOutput(const QSize& size)
	: mImg(size, QImage::Format_ARGB32)
{
}

/* misc
 ******************************************************************/
static float poly_dist(const Point3& pt, const QPolygonF& poly)
{
	// OMG!  So slow!  Probably should store line segments in an
	// easy-to-search format
	const int			size = poly.size();
	Segment3			seg;
	seg.set(poly[size-1], poly[0]);
	float				min = seg.distanceTo(pt);
	for (int k=1; k<size; k++) {
		seg.set(poly[k-1], poly[k]);
		const float		d = seg.distanceTo(pt);
		if (d < min) min = d;
	}
	return min;
}
