#ifndef FXDEFS_H
#define FXDEFS_H

#include <QImage>
#include <QPolygonF>

#include "collections/recyclelist.h"
#include "engine/archead.h"
#include "engine/cell.h"
#include "param/paramurl.h"

class Fx;

namespace fx {

// For visting
static const int		CONTINUE = 1;
static const int		STOP = 0;

/* VISIT-PARAMS
 ******************************************************************/
struct VisitParams {
	int					mDepth;

	// The depth that contains the layer FX
	static const int	LAYER_DEPTH = 0;
	// The depth of each layer's top chain
	static const int	ROOT_DEPTH = LAYER_DEPTH + 1;

	VisitParams();
};

/* VISITOR
 * For walking the Fx graph.
 ******************************************************************/
class Visitor {
public:
	virtual ~Visitor()	{ }

	// Answer CONTINUE, etc.
	virtual int		visit(Fx&, VisitParams&) = 0;

protected:
	Visitor()			{ }
};

/* WORK-ARGS.  Arguments to work().
 ******************************************************************/
class WorkArgs {
public:
	static const int	DISTANCE_CACHED = (1<<0);

public:
	WorkArgs();
	~WorkArgs();

	// Source polygon, if any
	QPolygonF			mShape;

	// Make sure buffers are allocated and cleared.
	bool				begin(const QSize&);

	// Get cached geometry info
	Cell*				getCell();
	// Get the current working image.  Image data stored as floats
	// with values of 0-1 in RGBA order, alpha premultiplied.
	float*				getImg();

	// Information about what types of data are currently cached.
	int					getCached() const;
	bool				isCached(const int flag) const;
	void				setCached(const int flag, const bool);
	void				uncacheAll();

	// This verifies that the distance is cached, and caches it if
	// it isn't.  Distance is in pixel values, with negative values
	// indicating outside the current shape, and positive values inside.
	// There shouldn't be an error possible, since the args are guaranteed
	// to be the output size.  Optionally answer the min and max values.
	void				cacheDistance(float* min = NULL, float* max = NULL);

	void				finish(QImage&);

	// Clients can request buffers of data for operating on.  Note
	// that if a Mask is returned, it's been verified to be the correct
	// size, so clients don't need to double check.
	Mask*				acquireMask(const int size);
	void				releaseMask(Mask*);

private:
	QSize				mSize;
	CellMask			mCell;
	Mask				mImg;
	int					mCached;
	// Distance info
	float				mMinD, mMaxD;

	RecycleList<Mask*>	mScratch;
};

/* WORK-OUTPUT.  Store the result of work processing.
 ******************************************************************/
class WorkOutput {
public:
	WorkOutput(const QSize&);

	// A buffer to store our final pixels
	QImage				mImg;
};

/* inline
 ******************************************************************/
inline int WorkArgs::getCached() const {
	return mCached;
}

inline bool WorkArgs::isCached(const int flag) const {
	return (mCached&flag) != 0;
}

} // fx

#endif // FXDEFS_H
