#include "surfaceneighbor.h"

#include "math/geo.h"

SurfaceNeighbor::SurfaceNeighbor()
	: mS(NULL)
	, mSize(0)
{
}

SurfaceNeighbor::~SurfaceNeighbor()
{
	realloc(mS, 0);
}

bool SurfaceNeighbor::init(const QuadList& l)
{
	mSize = 0;

	const int		size = l.size();
	if (size < 1) return true;
	Surface*		newS = (Surface*)realloc(mS, sizeof(Surface)*size);
	if (newS == NULL) return false;

	mS = newS;
	mSize = size;

	const QuadList::Entry*		e = l.head();
	Normal						n;
	int							i = 0;
	while (e != NULL) {
		mS[i].set(e->mQ);
		n.setOn(e->mQ, mS[i].mNormal);

		i++;
		e = e->mNext;
	}
	// Cache the normal info from all adjacent surfaces.
	fillAdjacentNormals();
	return true;
}

const Surface* SurfaceNeighbor::find(const float x, const float y) const
{
	for (int k=0; k<mSize; k++) {
		if (mS[k].containsCCW(x, y)) return &mS[k];
	}
	return NULL;
}

void SurfaceNeighbor::beginRow(const Surface& s, const float y)
{
	mA.y = mB.y = y;
	mA.x = s.mL-1;
	mB.x = s.mR+1;

	mZ[0] = 0;
	mZ[1] = 0;
	mX[0] = 1000000;
	mX[1] = -1000000;

	double				minY = 0, maxY = 0;
	int					minK = -1, maxK = -1;
	const Point3*		prev = &(s.mQ.mPt[Quad::SIZE-1]);
	const bool			isTri = s.mQ.mPt[1].sameAs2(s.mQ.mPt[2]);
	for (int k=0; k<Quad::SIZE; k++) {
		const Point3*	next = &(s.mQ.mPt[k]);
		// Handle the case where the quad is really a tri by skipping
		// one of the interior points.
		if (!(isTri && k == 2)) {
			if (geo::lineSegmentIntersection(mA, mB, *prev, *next, &mOut)) {
				if (mOut.x < mX[0]) {
					mX[0] = mOut.x;
					minY = mOut.y;
					minK = k;
				}
				if (mOut.x > mX[1]) {
					mX[1] = mOut.x;
					maxY = mOut.y;
					maxK = k;
				}
			}
		}
		prev = next;
	}
	if (minK < 0) {
//		System.out.println("Error in shader y=" + y + " q=" + s.mQ);
		return;
	}
	// Interpolate the left and right sides
	fill(s, minK, minY, 0);
	fill(s, maxK, maxY, 1);
}

void SurfaceNeighbor::getAt(const float x, Params& out) const
{
	const float			pos = (x-mX[0]) / (mX[1]-mX[0]);
	out.mZ = (mZ[0] * (1-pos)) + (mZ[1] * pos);
	out.mN.x = (mN[0].x * (1-pos)) + (mN[1].x * pos);
	out.mN.y = (mN[0].y * (1-pos)) + (mN[1].y * pos);
	out.mN.z = (mN[0].z * (1-pos)) + (mN[1].z * pos);
}

void SurfaceNeighbor::fill(const Surface& s, const int idx, const float y, const int dst)
{
	const int			pidx = (idx <= 0 ? Quad::SIZE-1 : idx-1);
	const Point3&		prv = s.mQ.mPt[pidx];
	const Point3&		nxt = s.mQ.mPt[idx];
	const float			segD = geo::dist(prv.x, prv.y, nxt.x, nxt.y),
						ptD = geo::dist(prv.x, prv.y, mX[dst], y);
	const float			pos = ptD / segD;
	// Depth
	mZ[dst] = ((prv.z * (1-pos)) + (nxt.z * pos));
	// Normal
	const Point3&		prev = s.mPtNormal[pidx];
	const Point3&		next = s.mPtNormal[idx];
	mN[dst].x = ((prev.x * (1-pos)) + (next.x * pos));
	mN[dst].y = ((prev.y * (1-pos)) + (next.y * pos));
	mN[dst].z = ((prev.z * (1-pos)) + (next.z * pos));
}

void SurfaceNeighbor::fillAdjacentNormals()
{
	for (int is=0; is<mSize; is++) {
		Surface&	src = mS[is];
		for (int k=0; k<Quad::SIZE; k++) {
			mC[k] = 1;
	//		src.mPtNormal[k].set(0, 0, 0);
			src.mPtNormal[k].set(src.mNormal);
		}
		for (int id=0; id<mSize; id++) {
			if (is == id) continue;
			// Right adjacent...
			fillAccum(src, 0, 1, mS[id], 3, 2);
			// Left adjacent...
			fillAccum(src, 3, 2, mS[id], 0, 1);
		}
/*
		for (int id=0; id<mSize; id++) {
			const Surface&		dst = mS[id];
			// Each point that's in common adds
			for (int srcK=0; srcK<Quad::SIZE; srcK++) {
				for (int dstK=0; dstK<Quad::SIZE; dstK++) {
					if (src.mQ.mPt[srcK].similarTo2(dst.mQ.mPt[dstK])) {
						src.mPtNormal[srcK] += dst.mNormal;
						c[srcK]++;
					}
				}
			}
		}
*/
		for (int k=0; k<Quad::SIZE; k++) {
			if (mC[k] > 1) src.mPtNormal[k] /= (float)(mC[k]);
		}
	}
}

void SurfaceNeighbor::fillAccum(	Surface& src,
									const int srcA, const int srcB,
									const Surface& dst,
									const int dstA, const int dstB)
{
	if (src.mQ.mPt[srcA].similarTo2(dst.mQ.mPt[dstA])
			&& src.mQ.mPt[srcB].similarTo2(dst.mQ.mPt[dstB])) {
		src.mPtNormal[srcA] += dst.mNormal;
		mC[srcA]++;
		src.mPtNormal[srcB] += dst.mNormal;
		mC[srcB]++;
	}
}
