#include "bevel.h"

#include <QtDebug>
#include <QList>
#include <QPainter>
#include <QPolygonF>
#include <QTransform>

#include <math.h>
#include "io/svgreader.h"
#include "math/geo.h"
#include "math/quadclipper.h"
#include "math/quadlist.h"
#include "math/report.h"
#include "param/param.h"
#include "fx/light.h"

using namespace fx;

static const bool			WIREFRAME = true;

static const QString		FX_NAME("Bevel");
static const QString		AMOUNT_NAME("Amount");
static const int			AMOUNT_ID = 0;
static const QString		LIGHT_ANGLE_NAME("Light");
static const int			LIGHT_ANGLE_ID = 20;
// Should be part of light
static const QString		CLR_NAME("Color");
static const int			CLR_ID = 100;		// XXX Shared with light.cpp
static const QString		ARC_NAME("Arc");
static const int			ARC_ID = 101;		// XXX Shared with light.cpp
// For testing...
static const QString		WIRE_NAME("Wire");
static const int			WIRE_ID = 21;

// Hardcoded limit on depth amount right now
static const float			DEPTH_MAX = 200;

#if 0
#define TEST_FILTER			(1)
//#define TEST_NO_CLIPPING	(1)
#endif
#ifdef TEST_FILTER
static void test_filter(QuadList& list);
#endif

// Extrusion
static bool extrude(const QPolygonF& src, const float depth, const float height,
					QuadList& out);
static void get_bevel_pt(	const Point3& a, const Point3& b, const Point3& c,
							const float depth, Point3& out);
static float directioned_dist(	const float ax, const float ay,
								const float bx, const float by,
								const float dist);
static bool isClockWise(const QPolygonF&);

#ifdef STORE_BISECTORS
static void drawPoint(QPainter& painter, QPen& pen, Point3& pt, const float min, const float max)
{
	int					c = (int)(((pt.z - min) / (max-min)) * 255.0f);
	if (c < 0) c = 0; else if (c > 255) c = 255;
	pen.setWidth(4);
	pen.setColor(QColor(c, c, c, 255));
	painter.setPen(pen);
	painter.drawLine(QPointF(pt.x, pt.y), QPointF(pt.x, pt.y));
}
#endif

/* BEVEL
 ******************************************************************/
Bevel::Bevel(Context& ctx)
	: Fx(ctx.nextFxUniqueId())
	, mDepth(80)
	, mLightAngle(0.35f)
	, mWire(0)
	, mLight(ctx)
{
#ifdef STORE_BISECTORS
	mWire = 0.35f;
#endif
	mLight.mClr.setId(CLR_ID);
	mLight.mArc.setId(ARC_ID);
	mLight.setGlobal(mLightAngle);
}

Bevel::Bevel(const Bevel& o)
	: Fx(o)
	, mDepth(o.mDepth)
	, mLightAngle(o.mLightAngle)
	, mWire(o.mWire)
	, mLight(o.mLight)
{
	mLight.mClr.setId(CLR_ID);
	mLight.mArc.setId(ARC_ID);
	mLight.setGlobal(mLightAngle);
}

Bevel::~Bevel()
{
}

Bevel* Bevel::copy() const
{
	return new Bevel(*this);
}

QString Bevel::getLabel() const
{
	return FX_NAME;
}

void Bevel::getParams(param::List& list, param::GetArgs& get) const
{
	list.addParam(&FX_NAME, AMOUNT_NAME, AMOUNT_ID, mDepth / DEPTH_MAX, get);
	list.addParam(&FX_NAME, LIGHT_ANGLE_NAME, LIGHT_ANGLE_ID, mLightAngle, get);
	list.addParam(&FX_NAME, WIRE_NAME, WIRE_ID, mWire, get);

	// Should be in light, but since that's not actually part of the
	// graph and won't be visited, needs to be here for now
//	if (!mLight.getParams(list, filter)) ans = false;
	if (get.supports(param::COLOR_FEATURE)) {
		list.addTypeParam(CLR_NAME, CLR_ID, param::COLOR_BROWSER_TYPE, get);
	}
	mLight.mClr.getParams(list, get);

	if (get.supports(param::ARC_FEATURE)) {
		list.addTypeParam(ARC_NAME, ARC_ID, param::ARC_BROWSER_TYPE, get);
	}
	mLight.mArc.getParams(list, get);
}

param::Parent* Bevel::findParent(param::UrlIterator& it)
{
	// If I hit a composite object, there are two possibilities:
	// If the URL is a leaf, then someone is trying to replace
	// the entire composite.  Otherwise they're looking to set a param.
	if (it.matches(CLR_ID)) {
		if (it.isLeaf()) return this;
		++it;
		return mLight.mClr.findParent(it);
	} else if (it.matches(ARC_ID)) {
		if (it.isLeaf()) return this;
		++it;
		return mLight.mArc.findParent(it);
	}
	return this;
}

void Bevel::setParam(const param::UrlIterator& it, param::SetArgs& sa)
{
	if (it.matches(AMOUNT_ID)) {
		float		trans = sa.mV * DEPTH_MAX;
		if (trans < 0) trans = 0;
		else if (trans > DEPTH_MAX) trans = DEPTH_MAX;
		if (mDepth == trans) return;		// no change
		mDepth = trans;
	} else if (it.matches(LIGHT_ANGLE_ID)) {
		if (mLightAngle == sa.mV) return;	// no change
		mLightAngle = sa.mV;
		mLight.setGlobal(mLightAngle);
	} else if (it.matches(WIRE_ID)) {
		if (mWire == sa.mV) return;			// no change
		mWire = sa.mV;
	// If I reached here, then I am a direct set.  Any parameters
	// inside the arc would go directly there
	} else if (it.matches(CLR_ID)) {
		mLight.mClr.swap(sa.mClr);
	} else if (it.matches(ARC_ID)) {
		mLight.mArc.setContentTo(sa.mArc);
	} else {
		return;
	}
	sa.mChanged = true;
}

void Bevel::work(fx::WorkArgs& arg, fx::WorkOutput& out)
{
//		qDebug() << "BEVEL DEPTH=" << mDepth << " input size=" << f.mShape.size();
	if (arg.mShape.size() < 1) return;

	// Extrude the polygon
	QuadList			extruded;
	if (!extrude(arg.mShape, mDepth, mDepth, extruded)) return;

#ifdef TEST_FILTER
	test_filter(extruded);
#endif

	// Clip the overlaps
	const float			DEPTH_SCALE = 1;
	QuadList			clipped;
#ifdef TEST_NO_CLIPPING
	extruded.swap(clipped);
	clipped.scaleToZ(mDepth * DEPTH_SCALE);
#else
	QuadClipper			clipper;
	clipper.on(extruded, mDepth * DEPTH_SCALE, clipped);
#endif
#if 1
//qDebug() << "" << clipped;

	// Draw the polygon -- technically this should happen in a shape FX
	{
		QPainter		painter(&out.mImg);
		QBrush			brush;
		brush.setColor(QColor(120, 120, 120));
		brush.setStyle(Qt::SolidPattern);

		painter.setPen(Qt::NoPen);
		painter.setBrush(brush);
		painter.setRenderHint(QPainter::Antialiasing);
		painter.drawPolygon(arg.mShape);
	}

	// Built in for now, maybe chainable some day
	if (!mLight.set(clipped)) return;
//	light.setGlobal(mLightAngle);
	mLight.work(arg, out);

#if 0
	qDebug() << "BEVEL OUTPUT";
	QuadList::Entry*	e = clipped.head();
	while (e != NULL) {
		qDebug() << "\te=" << e->mQ;
		e = e->mNext;
	}
#endif


	if (mWire > 0) {
		QPainter		painter(&out.mImg);
		QPen			pen(QColor(0, 0, 0, (int)(mWire*255)));
		pen.setWidth(2);
		painter.setPen(pen);
		painter.setBrush(Qt::NoBrush);
		painter.setRenderHint(QPainter::Antialiasing);

		const QuadList::Entry*		e = clipped.head();
		while (e != NULL) {
			const Point3*			prev = &(e->mQ.mPt[Quad::SIZE-1]);
			for (int i=0; i<Quad::SIZE; i++) {
				const Point3*		next = &(e->mQ.mPt[i]);
				painter.drawLine(QPointF(prev->x, prev->y), QPointF(next->x, next->y));
				prev = next;
			}
			e = e->mNext;
		}

#ifdef STORE_BISECTORS
#ifdef REPORT_GEO
		// Draw bisection lines
//			pen.setColor(QColor(0, 255, 0, (int)(mWire*255)));
		pen.setColor(QColor(0, 255, 0, 255));
		painter.setPen(pen);
		float					minZ = 1000000, maxZ = -1000000;
		for (int i=0; i<clipper.mReport.mSeg.size(); i++) {
			const Point3&		a = clipper.mReport.mSeg[i].mPt[0];
			const Point3&		b = clipper.mReport.mSeg[i].mPt[1];
			painter.drawLine(QPointF(a.x, a.y), QPointF(b.x, b.y));
			if (a.z < minZ) minZ = a.z;
			if (b.z < minZ) minZ = b.z;
			if (a.z > maxZ) maxZ = a.z;
			if (b.z > maxZ) maxZ = b.z;
		}
//	qDebug() << "MIN Z=" << minZ << " MAX Z=" << maxZ;
		// Draw bisection points based on Z value
//			pen.setColor(QColor(0, 255, 0, (int)(mWire*255)));
		pen.setColor(QColor(0, 255, 0, 255));
		painter.setPen(pen);
		for (int i=0; i<clipper.mReport.mSeg.size(); i++) {
			drawPoint(painter, pen, clipper.mReport.mSeg[i].mPt[0], minZ, maxZ);
			drawPoint(painter, pen, clipper.mReport.mSeg[i].mPt[1], minZ, maxZ);
		}
#endif // REPORT_GEO
#endif // STORE_BISECTORS
	}
#endif
}

/* Misc
 ********************************************/
static void conv(const QPointF& src, Point3& dst)
{
	dst.x = src.x();
	dst.y = src.y();
}

static bool similar(const QPointF& a, const QPointF& b)
{
	return fabs(a.x()-b.x()) < 0.1f && fabs(a.y()-b.y()) < 0.1f;
}

// Convert the polygon into extruded, 3d quads
static bool extrude(const QPolygonF& src, const float depth, const float height,
					QuadList& out)
{
	const bool		isClosed = similar(src[0], src[src.size()-1]);

	// Reverse the order of iteration for opposite effects
	const int		min = 0, max = src.size()-1 + (isClosed ? -1 : 0);
	int				i = 0, stop = max + 1, inc = 1;
	if (stop <= i) return false;
	// If the points are ordered clockwise, iterate backwards.
	if (isClockWise(src)) {
		i = max;
		stop = -1;
		inc = -1;
	}
	int				pi = i - inc,
					ni = i + inc,
					nni = i + inc + inc;
	Point3			inter, p1, p3,
					prv(0, 0, 0), cur(0, 0, 0),
					nxt(0, 0, 0), nxtNxt(0, 0, 0);

	while (i != stop) {
		// Constrain the indexes to valid values
		if (pi > max) pi = 0;
		else if (pi < min) pi = max;
		if (ni > max) ni = 0;
		else if (ni < min) ni = max;
		if (nni > max) nni = 0;
		else if (nni < min) nni = max;

		conv(src[pi], prv);
		conv(src[i], cur);
		conv(src[ni], nxt);
		conv(src[nni], nxtNxt);

		// Every 2 points are extruded into a new quad, by generating
		// a new point for each.
		get_bevel_pt(prv, cur, nxt, depth, p1);
		get_bevel_pt(cur, nxt, nxtNxt, depth, p3);

		p1.z = height;
		p3.z = height;

		// If the quad will interesect itself, that becomes
		// the new interior.  This turns it into a fake triangle.
		if (geo::lineSegmentIntersection(cur, p1, nxt, p3, &inter)) {
			// We have to scale the height correctly based on the clamp
			inter.z = height*( (inter.dist3(cur) / p1.dist3(cur)));
			p1.set(inter);
			p3.set(inter);
		}

		// The height is simply the distance from the origin point.
//		p1.z = p1.dist2(cur);
//		p3.z = p3.dist2(nxt);

		// The points are ordered counter clockwise, so the new
		// p1 is the first interior, and p3 is the second.
		out.push(cur, p1, p3, nxt);

		// Increment
		pi = i;
		i += inc;
		ni += inc;
		nni += inc;
	}

	return true;
}

// Given three ordered points (a, b, c) and a depth, find
// a new point representing the bevel at the given depth.
static void get_bevel_pt(	const Point3& a, const Point3& b, const Point3& c,
							const float depth, Point3& out)
{
	const float			m_ab = geo::perpendicularSlope(a.x, a.y, b.x, b.y),
						m_bc = geo::perpendicularSlope(b.x, b.y, c.x, c.y);

	// Use the segment direction to decide which direction
	// to extend the new point in.
	const float			dist_ab = directioned_dist(a.x, a.y, b.x, b.y, depth),
						dist_bc = directioned_dist(b.x, b.y, c.x, c.y, depth);

	Point3				ab1, ab2, bc1, bc2;

	// Get two new line segments, perpendicular to the supplied points.
	geo::newPtOnLine(a.x, a.y, m_ab, dist_ab, ab1);
	geo::newPtOnLine(b.x, b.y, m_ab, dist_ab, ab2);
	geo::newPtOnLine(b.x, b.y, m_bc, dist_bc, bc1);
	geo::newPtOnLine(c.x, c.y, m_bc, dist_bc, bc2);

	// Find the interesection of these lines.
	geo::lineIntersection(ab1, ab2, bc1, bc2, out);
}

static float directioned_dist(	const float ax, const float ay,
								const float bx, const float by,
								const float dist)
{
	double				deg = geo::degree(bx-ax, by-ay);
	// My getDegree func doesn't correspond exactly to the
	// angles that determine direction.
	deg -= 90;
	if (deg < 0) deg += 360;
	if (deg >= 0 && deg < 180) return -dist;
	return dist;
}

// Answer true if the polygon is ordered clockwise, false
// if it's counter.  First and last point can't repeat.
// Original function by Paul Bourke
// http://debian.fmi.uni-sofia.bg/~sergei/cgsr/docs/clockwise.htm
static bool isClockWise(const QPolygonF& p)
{
	if (p.size() < 2) return false;
	const bool		isClosed = similar(p[0], p[p.size()-1]);
	const int		size = p.size() + (isClosed ? -1 : 0);

	if (size < 3) return false;

	int				count = 0;
	for (int i=0;i<size;i++) {
		const int	j = (i + 1) % size;
		const int	k = (i + 2) % size;
		float		z  = (p[j].x() - p[i].x()) * (p[k].y() - p[j].y());
		z -= (p[j].y() - p[i].y()) * (p[k].x() - p[j].x());
		if (z < 0) count--;
		else if (z > 0) count++;
	}
	if (count > 0) return false;
	else if (count < 0) return true;
	else return true;		// technically undetermined
}

#ifdef TEST_FILTER
static void test_filter(QuadList& list)
{
	QuadList::Entry*		e = list.detach();
	int						idx = 0;
	while (e != NULL) {
		QuadList::Entry*	next = e->mNext;
		e->mNext = NULL;

		if (idx == 1 || idx == 10) {
//			qDebug() << "PROCESS=" << e->mQ;
			list.addTail(e);
		} else delete e;

		e = next;
		idx++;
	}
}
#endif
