#include "light.h"

#include <stdlib.h>
#include <math.h>
#include <QtDebug>
#include "engine/arc.h"
#include "engine/colorf.h"
#include "engine/surface.h"
#include "math/geo.h"

using namespace fx;

static const QString		FX_NAME("Light");

static const int			CLR_ID = 100;		// XXX Shared with bevel.cpp
static const int			ARC_ID = 101;		// XXX Shared with bevel.cpp

/* LIGHT
 ******************************************************************/
Light::Light(Context& ctx)
	: Fx(ctx.nextFxUniqueId())
	, mGlobalAngle(0)
	, mInited(false)
{
	ctx.getColorSrc(mClr, COLOR_SRC_DUO);
	ctx.getArc(mArc, 3);
}

Light::Light(const Light& o)
	: Fx(o)
	, mGlobalAngle(o.mGlobalAngle)
	, mClr(o.mClr)
	, mArc(o.mArc)
	, mInited(false)
{
}

Light::~Light()
{
}

Light* Light::copy() const
{
	return new Light(*this);
}

static float max_z(const Quad& q)
{
	float		z = q.mPt[0].z;
	for (int i=1; i<Quad::SIZE; i++) {
		if (q.mPt[i].z > z) z = q.mPt[i].z;
	}
	return z;
}

bool Light::set(const QuadList& l)
{
	mInited = mShader.init(l);
	if (!mInited) return false;

	mMaxZ = 0;
	const QuadList::Entry*		e = l.head();
	while (e != NULL) {
		mMaxZ = fmax(mMaxZ, max_z(e->mQ));

		e = e->mNext;
	}

	return true;
}

void Light::setGlobal(const float angle)
{
	mGlobalAngle = angle;
}

QString Light::getLabel() const
{
	return FX_NAME;
}

void Light::getParams(param::List&, param::GetArgs&) const
{
/*
	bool			ans = true;
	if (f != NULL && f->supports(Param::Filter::ARC_FEATURE)) {
		add_arc_browser_param(list, ARC_NAME, getId(), ARC_ID, f, &ans);
	}
	return ans;
*/
}

/*
bool Light::setParam(fx::ParamArgs& p)
{
	if (p.mKey.mParam == ARC_ID) {
		mArc.setTo(p.mArc);
		return true;
	}
	return false;
}
*/
param::Parent* Light::findParent(param::UrlIterator& it)
{
	if (!it.hasValue()) return NULL;
	return NULL;
}

void Light::setParam(const param::UrlIterator&, param::SetArgs&)
{
}

static inline float lightFrom(const Point3& normal, const Point3& light, const float length)
{
	return (normal.x*light.x + normal.y*light.y + normal.z*light.z) / (normal.length() * length);
}

static void get_circle_edge(	const Point3& cen, const float angle, const float dist,
								Point3& out)
{
	float				degree = angle * 360;
	if (degree < 0) degree += 360;
	else if (degree > 1) degree -= 360;
	degree *= geo::DTOR;
	out.x = (cen.x + (cos(degree) * dist));
	out.y = (cen.y - (sin(degree) * dist));
	// Vary this for some nice effects
	out.z = 0;
}

void Light::work(fx::WorkArgs& arg, fx::WorkOutput& out)
{
	if (!mInited) return;

	const QSize			imgSize = out.mImg.size();
	float*				img = arg.getImg();
	const ColorSrc*		csrc = mClr.getSrc();
	if (img == NULL || csrc == NULL) return;
	// Set the global location based on my angle and size.
	const Point3		cen(float(imgSize.width())/2, float(imgSize.height())/2, 0);
	Point3				global;		// light
	get_circle_edge(cen, mGlobalAngle, (cen.x+cen.y)*1.5f, global);

	const int			w = imgSize.width(), h = imgSize.height();
	const Surface*		s = NULL;
	Point3				pt;
	const float			globalLength = global.length(),
						maxZ = (mMaxZ <= 0 ? 1 : mMaxZ);
	int					pix = 0;

	if (!mArcP.begin(&mArc, w, arc::HAS_DISTANCE_SRC | arc::HAS_NORMAL_SRC)) return;

	// The final values are written to a float row, in case I have
	// an arc that needs to do processing.
	float*				rowV = mRowV.get(w);
	float*				rowMask = mRowMask.get(w);
	if (rowV == NULL || rowMask == NULL) return;

	Arc*				arc = mArc.getArc();
	color_f				clr;

	for (int y=0; y<h; y++) {
		pt.y = y;

		set_to_neg_1(rowMask, w);

		// Find the light and distance values for each row
		for (int x=0; x<w; x++) {
			pt.x = x;

			if (s == NULL || !s->containsCCW(pt.x, pt.y)) {
				if ((s = mShader.find(pt.x, pt.y)) != NULL) {
					mShader.beginRow(*s, pt.y);
				}
			}

			if (s != NULL) {
				// XXX Should set antialias value.
				rowMask[x] = 1;
				mShader.getAt(pt.x, mShaderP);
				mArcP.mNormalRow[x] = lightFrom(mShaderP.mN, global, globalLength);
				mArcP.mDistanceRow[x] = (mShaderP.mZ / maxZ);
			}
		}

		// Process the row
		set_to_1(rowV, w);
		if (arc != NULL) arc->work(mArcP, rowV);

		// Apply the row
		for (int x=0; x<w; x++) {
			if (rowMask[x] > 0 && rowV[x] != 0) {
				// XXX Currently it's -1 to 1 but needs to be 0 to 1
				rowV[x] = (rowV[x]+1) * 0.5f;

				csrc->get(rowV[x], clr);
				if (clr.a > 0) srcOver(clr, img, pix);
			}
			pix++;
		}
	}
}
