#include "shapesrc.h"

#include <QDebug>

#include "io/svgreader.h"
#include "fxui/shapebrowser.h"

using namespace fx;

static const QString		FX_NAME("Shape");
static const QString		FN_NAME("Filename");
static const int			FN_ID = 0;

static bool load_poly_from_svg(const QString& filename, const QRectF& dstF,
								QPolygonF& dst);

/* SHAPE-SRC
 ******************************************************************/
ShapeSrc::ShapeSrc(Context& ctx)
	: Fx(ctx.nextFxUniqueId())
	, mNeedsCache(0)
{
}

ShapeSrc::ShapeSrc(const ShapeSrc& o)
	: Fx(o)
	, mFilename(o.mFilename)
	, mNeedsCache(0)
{
}

ShapeSrc* ShapeSrc::copy() const
{
	return new ShapeSrc(*this);
}

QString ShapeSrc::getLabel() const
{
	return FX_NAME;
}

void ShapeSrc::getParams(param::List& list, param::GetArgs& get) const
{
	list.addParam(&FX_NAME, FN_NAME, FN_ID, mFilename, get);
}

param::Parent* ShapeSrc::findParent(param::UrlIterator& it)
{
	if (!it.isLeaf()) return NULL;
	return this;
}

void ShapeSrc::setParam(const param::UrlIterator& it, param::SetArgs& sa)
{
	if (it.matches(FN_ID)) {
		if (mFilename == sa.mStr) return;		// no change
		mFilename = sa.mStr;
		mNeedsCache = fx::WorkArgs::DISTANCE_CACHED;
	}
	sa.mChanged = true;
}

void ShapeSrc::work(fx::WorkArgs& arg, fx::WorkOutput& o)
{
	if (mNeedsCache != 0) arg.setCached(mNeedsCache, false);
	mNeedsCache = 0;

	// Load the polygon
	arg.mShape.clear();
	const QSize			size = o.mImg.size();
	// Hard code a destination frame that gives a little room for a shadow
	const float			edge = 0.05f;
	const float			w = size.width(), h = size.height();
	const float			l = w*edge, t = h*edge, r = w*(1-edge), b = h*(1-edge);
	load_poly_from_svg(mFilename, QRectF(l, t, r-t, b-t), arg.mShape);
}

FxUi* ShapeSrc::newUi(FxUi::Create& c) const
{
	param::Url		url;
	url.prepend(FN_ID);
	Parent::makeUrl(this, url);
	return new fxui::ShapeBrowser(c, url);
}

/* Misc
 ********************************************/
static bool load_poly_from_svg(const QString& filename, const QRectF& dstF,
								QPolygonF& dst)
{
	// Load the path from the file
	SvgReader			r;
	QPainterPath		p;
	if (!r.load(filename, &p)) return false;

	// Conver the path to a polygon
	QList<QPolygonF>	pl = p.toSubpathPolygons();
	if (pl.size() < 1) return false;

	const QPolygonF&	poly = pl.at(0);
	QPolygonF			newPoly;

	// Translate the polygon to local coords
	const QRectF		srcF(poly.boundingRect());
	QTransform			trans;
	{
		QPolygonF		a, b;
		a << srcF.topLeft() << srcF.topRight() << srcF.bottomRight() << srcF.bottomLeft();
		b << dstF.topLeft() << dstF.topRight() << dstF.bottomRight() << dstF.bottomLeft();
		if (!QTransform::quadToQuad(a, b, trans)) return false;
	}
	for (int k=0; k<poly.size(); k++) {
		dst.append( poly.at(k) * trans);
	}
	return true;
}
