#include "fxui/parameditor.h"

#include <assert.h>

#include <QDebug>
#include <QPainter>

#include "engine/engineparam.h"
#include "model/context.h"
#include "ui/look.h"
#include "fxui/arcbrowser.h"
#include "fxui/colorbrowser.h"

using namespace fxui;

/* PARAM-EDITOR::PARAM-FILTER
 ******************************************************************/
class ParamEditor::ParamFilter : public param::Filter
{
public:
	virtual bool accept(const QString*, const QString*, const int type) const
	{
		return type == Param::FLOAT_TYPE || type == param::ARC_BROWSER_TYPE || type == param::COLOR_BROWSER_TYPE;
	}

	virtual bool supports(const int feature) const
	{
		return (feature&(param::ARC_FEATURE | param::COLOR_FEATURE)) != 0;
	}

};
static const ParamEditor::ParamFilter		gFilter;

/* PARAM-EDITOR
 ******************************************************************/
ParamEditor::ParamEditor(Create& c, const param::Url& url)
	: mPrj(c.mPrj)
	, mArcs(c.mArcs)
	, mClrs(c.mClrs)
	, mUrl(url)
	, mNeedsCache(true)
{
}

ParamEditor::~ParamEditor()
{
	uncache();
}

void ParamEditor::setParent(Parent* p, const Look& look)
{
	assert(p != NULL);
	inherited::setParent(p, look);

	// Should check to see if the folder date stamp changed
	if (mNeedsCache) {
		cache(p->entryH(), look);
		mNeedsCache = false;
	}

	for (int i=0; i<mS.size(); i++) {
		if (mS[i] == NULL) continue;

		if (mS[i]->mBrowser != NULL) p->addBrowser(mS[i]->mBrowser);
		else p->add(mS[i]);
	}
}

int ParamEditor::onProjectChanged()
{
	if (!mPrj.contains(mUrl)) return FxUi::UI_ERROR;
	uncache();
	mNeedsCache = true;
	return FxUi::UI_CHANGED;
}

void ParamEditor::cache(const int entryH, const Look&)
{
	param::List				list;
	param::GetArgs			ga;
	ga.mUrl = mUrl;
	ga.mFilter = &gFilter;
	mPrj.getParams(list, ga);

	param::ListIterator		it(list);
	const Param*			p;
	QString					groupName;
	FxE*					lastE = NULL;
	while ((p = *it) != NULL) {
		int					flags = 0;
		if (groupName == p->mGroupName) {
			if (lastE != NULL) lastE->mFlags |= Entry::CONNECT_RIGHT;
			flags = Entry::CONNECT_LEFT;
		}
		groupName = p->mGroupName;
		FxE*				e = new FxE(*this, *(p), QSize(60, entryH), flags);
		if (e != NULL) {
			if (p->mType == param::ARC_BROWSER_TYPE) {
				e->mBrowser = new ArcBrowser(mPrj, mArcs, *p, QSize(entryH, entryH));
				if (e->mBrowser != NULL) {
					mS.append(e);
					lastE = e;
				} else delete e;
			} else if (p->mType == param::COLOR_BROWSER_TYPE) {
				e->mBrowser = new ColorBrowser(mPrj, mClrs, *p, QSize(entryH, entryH));
				if (e->mBrowser != NULL) {
					mS.append(e);
					lastE = e;
				} else delete e;
			} else {
				mS.append(e);
				lastE = e;
			}
		}

		++it;
	}
}

void ParamEditor::uncache()
{
	for (int i=0; i<mS.size(); i++) delete mS[i];
	mS.clear();
}

/* FX-E
 *******************************/
ParamEditor::FxE::FxE(ParamEditor& pe, const Param& p,
					  const QSize& size, const int flags)
	: mBrowser(NULL)
	, mFlags(flags)
	, mP(pe)
	, mSize(size)
	, mParamType(p.mType)
{
	mUrl = p.mUrl;
	mLabel = p.mParamName;
	mV = p.mFV;
}

ParamEditor::FxE::~FxE()
{
	delete mBrowser;
}

QSize ParamEditor::FxE::size() const
{
	return mSize;
}

bool ParamEditor::FxE::getParam(Param& p) const
{
	p.mUrl = mUrl;
	p.mParamName = mLabel;
	p.mType = mParamType;
	return true;
}

void ParamEditor::FxE::onPaint(paint_args& pa)
{
	QPainter&			painter = *pa.mPainter;
	const Look&			look = *pa.mLook;
	const QRect&		frame = pa.mFrame;

	painter.setRenderHint(QPainter::Antialiasing);

	// Panel background
	painter.setPen(Qt::NoPen);
	painter.setBrush(look.mParamBgB);
	// Right now do this the easy way -- assume both sides are
	// closed, fill in if not.
	painter.drawRoundedRect(frame, look.mRound, look.mRound);
	if ((mFlags&CONNECT_LEFT) != 0) painter.drawRect(frame.left(), frame.top(), look.mRound, frame.height());
	if ((mFlags&CONNECT_RIGHT) != 0) painter.drawRect(frame.right()-look.mRound, frame.top(), look.mRound+1, frame.height());

	painter.setBrush(look.mParamFgB);

	// Label
	if (mLabel.size() > 0) {
		painter.setPen(look.mParamFgC);
		painter.drawText(frame.left() + look.mPadX, frame.top() + look.mPadY + look.mFa, mLabel);
		painter.setPen(Qt::NoPen);
	}

	// Param value background
	const int		l = frame.left() + look.mPadX;
	const int		t = frame.top() + look.mPadY + look.mFh + look.mPadY;
	const int		valueW = frame.width() - look.mPadX * 2;
	const int		valueH = (frame.bottom() - look.mPadY) - t;
	painter.setBrush(look.mParamBgToneB);
	painter.drawRoundedRect(l, t, valueW, valueH, look.mRound, look.mRound);

	// Param value
	const int		curW = (int)(float(valueW) * mV);
	painter.setBrush(look.mParamFgB);
	painter.drawRoundedRect(l, t, curW, valueH, look.mRound, look.mRound);
}

void ParamEditor::FxE::mousePressEvent(const mouse_args& down)
{
	mP.mDownPt.setX(down.mEvent->x());
	mP.mDownPt.setY(down.mEvent->y());
	mP.mDownV = mV;
}

static float new_value(const QPoint& startPt, const int curX, const QRect& f, const float origV)
{
	const int		offset = curX - startPt.x();
	const float		newV = origV + (float(offset) / (float(f.width()*2)));
	if (newV <= 0) return 0;
	if (newV >= 1) return 1;
	return newV;
}

void ParamEditor::FxE::mouseMoveEvent(const mouse_args& ma)
{
	const float		newV = new_value(mP.mDownPt, ma.mEvent->x(), ma.mFrame, mP.mDownV);
	mP.mPrj.setParamF(mUrl, newV);
}

void ParamEditor::FxE::mouseReleaseEvent(const mouse_args&)
{
}

int ParamEditor::FxE::onParamChanged(const Param& p)
{
	if (mUrl == p.mUrl) {
		// Indicate that not only is there nothing to update,
		// but I can stop looking for something.
		if (mV == p.mFV) return NO_PARAMS;
		mV = p.mFV;
		return PARAM;
	}
	return WRONG_PARAM;
}
