#include "fxui/shapebrowser.h"

#include <assert.h>

#include "app/defs.h"
#include "app/environment.h"
#include "io/svgreader.h"
#include "model/context.h"
#include "ui/look.h"

using namespace fxui;

static bool transform_frame(	const QRectF& srcF, const int dstS, QTransform& trans,
								const Look&);
/* SHAPE-BROWSER
*******************************/
ShapeBrowser::ShapeBrowser(Create& c, const param::Url& fileUrl)
	: mEnv(c.mEnv)
	, mPrj(c.mPrj)
	, mFileUrl(fileUrl)
	, mNeedsCache(true)
{
}

ShapeBrowser::~ShapeBrowser()
{
	uncache();
}

void ShapeBrowser::setParent(Parent* p, const Look& look)
{
	assert(p != NULL);
	inherited::setParent(p, look);

	// Should check to see if the folder date stamp changed
	if (mNeedsCache) {
		cache(p->entryH(), look);
		mNeedsCache = false;
	}

	for (int i=0; i<mS.size(); i++) {
		if (mS[i] == NULL) continue;
		p->add(mS[i]);
	}
}

int ShapeBrowser::onProjectChanged()
{
	if (!mPrj.contains(mFileUrl)) return FxUi::UI_ERROR;
	return FxUi::UI_CHANGED;
}

void ShapeBrowser::cache(const int entryH, const Look& look)
{
	// Create all the entries here.
	// Of course, this should be in a separate thread
	// and there should be a signal to indicate when a
	// new entry is ready to display.  That separate thread
	// also needs an object that copies the relevant fiels
	// from "Look"

	const QDir			dir(mEnv.getInstallPath(SHAPES_DIR));
	const QString		pdir(dir.absolutePath());
	const QStringList	files = dir.entryList(QDir::Files);
	for (int k=0; k<files.size(); k++) {
		FxE*			s = new FxE(mPrj, mFileUrl, pdir + QDir::separator() + files.at(k));
		if (s != NULL && s->setImg(entryH, look)) {
			mS.append(s);
		} else delete s;
	}
}

void ShapeBrowser::uncache()
{
	for (int i=0; i<mS.size(); i++) delete mS[i];
	mS.clear();
}

/* FX-E
 *******************************/
ShapeBrowser::FxE::FxE(Context& prj, const param::Url& url, const QString& fn)
	: mPrj(prj)
	, mUrl(url)
	, mFilename(fn)
{
}

bool ShapeBrowser::FxE::setImg(const int size, const Look& look)
{
	// Load a path for the underlying SVG file.
	SvgReader		reader;
	QPainterPath	path;
	if (!reader.load(mFilename, &path)) {
		qDebug() << "ERROR LOADING " << mFilename;
	}

	QImage*			img = new QImage(size, size, QImage::Format_RGB888);
	if (img == NULL) return false;

	{
		QPainter		painter(img);
		// Clear the background
		painter.setPen(Qt::NoPen);
		painter.setBrush(look.mDashBgB); //.mAlphaB);
		painter.drawRect(0, 0, size, size);
		// Draw a param panel
		painter.setBrush(look.mParamBgB);
		painter.setRenderHint(QPainter::Antialiasing);
		painter.drawRoundedRect(0, 0, size, size, look.mRound, look.mRound);
		// Translate the path to local coords
		const QRectF	pathF(path.boundingRect());
		QTransform		trans;
		transform_frame(pathF, size, trans, look);
		painter.setTransform(trans);
		// Draw the path
		painter.setBrush(look.mParamFgB);
		painter.drawPath(path);
	}

	// Kinda weird -- where's the error check?
	mImg = QPixmap::fromImage(*img);
	delete img;
	return true;
}

QSize ShapeBrowser::FxE::size() const
{
	return mImg.size();
}

void ShapeBrowser::FxE::onPaint(paint_args& pa)
{
	pa.mPainter->drawPixmap(pa.mFrame.left(), pa.mFrame.top(), mImg);
}

void ShapeBrowser::FxE::mousePressEvent(const mouse_args&)
{
}

void ShapeBrowser::FxE::mouseMoveEvent(const mouse_args&)
{
}

void ShapeBrowser::FxE::mouseReleaseEvent(const mouse_args& down)
{
	if (mFilename.size() > 0 && down.mEntry != NULL) {
		mPrj.setParamStr(mUrl, mFilename);
	}
}

/* MISC
 *******************************/
static bool transform_frame(const QRectF& srcF, const int dstS, QTransform& trans, const Look& look)
{
	QPolygonF		a, b;
	const int		w = dstS - look.mPadX*2, h = dstS - look.mPadY*2;
	QRectF			dstF(look.mPadX, look.mPadY, w-1, h-1);
	a << srcF.topLeft() << srcF.topRight() << srcF.bottomRight() << srcF.bottomLeft();
	b << dstF.topLeft() << dstF.topRight() << dstF.bottomRight() << dstF.bottomLeft();

	return QTransform::quadToQuad(a, b, trans);
}
