#include "geo.h"

#include <QtGlobal>

#include "math.h"
#include "math/primitives/point3.h"

using namespace geo;

const float			geo::EPS = 0.00000000001f;
const float			geo::DTOR = 0.017453f;
const float			geo::RTOD = 57.2957795f;

float geo::slope(const float x0, const float y0, const float x1, const float y1)
{
	return (y1 - y0) / (x1 - x0);
}

float geo::perpendicularSlope(const float x0, const float y0, const float x1, const float y1) {
	return -((x1-x0) / (y1-y0));
}

float geo::dist(const float x0, const float y0, const float x1, const float y1) {
	return sqrt(((x0-x1)*(x0-x1)) + ((y0-y1)*(y0-y1)));
}

float geo::distSquared(const float x0, const float y0, const float x1, const float y1)
{
	return ((x0-x1)*(x0-x1)) + ((y0-y1)*(y0-y1));
}

void geo::lineIntersection(	const Point3& a1, const Point3& a2,
							const Point3& b1, const Point3& b2,
							Point3& out) {
	lineIntersection(a1.x, a1.y, a2.x, a2.y, b1.x, b1.y, b2.x, b2.y, out);
}

void geo::lineIntersection(	const float a1x, const float a1y,
							const float a2x, const float a2y,
							const float b1x, const float b1y,
							const float b2x, const float b2y,
							Point3& out) {
	const float			A1 = a2y - a1y,
						B1 = a1x - a2x;
	const float			C1 = (A1*a1x)+(B1*a1y);
	const float			A2 = b2y - b1y,
						B2 = b1x - b2x;
	const float			C2 = (A2*b1x)+(B2*b1y);
	// Do the algorithm
	const float			det = (A1*B2) - (A2*B1);
	if (det == 0) {
		// Parallel
		out.x = a2x;
		out.y = a2y;
	} else {
		out.x = (B2*C1 - B1*C2)/det;
		out.y = (A1*C2 - A2*C1)/det;
	}
}

bool geo::lineSegmentIntersection(	const Point3& a1, const Point3& a2,
									const Point3& b1, const Point3& b2,
									Point3* out) {
	return lineSegmentIntersection(a1.x, a1.y, a2.x, a2.y, b1.x, b1.y, b2.x, b2.y, out);
}

bool geo::lineSegmentIntersection(	const float x1, const float y1,
									const float x2, const float y2,
									const float x3, const float y3,
									const float x4, const float y4,
									Point3* out) {
	const float		bx = x2 - x1,
					by = y2 - y1,
					dx = x4 - x3,
					dy = y4 - y3;
	const float		b_dot_d_perp = bx * dy - by * dx;
	if (b_dot_d_perp == 0) {
		return false;
	}
	const float		cx = x3 - x1,
					cy = y3 - y1;
	const float		t = (cx * dy - cy * dx) / b_dot_d_perp;
	if (t < 0 || t > 1) {
		return false;
	}
	const float		u = (cx * by - cy * bx) / b_dot_d_perp;
	if (u < 0 || u > 1) {
		return false;
	}
	if (out != NULL) {
		out->x = x1+t*bx;
		out->y = y1+t*by;
	}
	return true;
}

float geo::areaOfTri(	const float ax, const float ay,
						const float bx, const float by,
						const float cx, const float cy)
{
	const float		a = (ax - cx),
					b = (ay - cy),
					c = (bx - cx),
					d = (by - cy);
	return 0.5f * fabs((a*d) - (b*c));
}

// Technically a function on a triangle, if we had one.
// This works by taking the given point, and converting the
// triangle into three triangles with the new point, then
// comparing the area of that with the area of the triangle.
bool geo::isInsideTri(	const float ax, const float ay,
						const float bx, const float by,
						const float cx, const float cy,
						const float x, const float y)
{
	const float		t = areaOfTri(ax, ay, bx, by, cx, cy);
	float			sum = areaOfTri(ax, ay, bx, by, x, y);
	if (sum > t) return false;

	sum += areaOfTri(bx, by, cx, cy, x, y);
	if (sum > t) return false;

	sum += areaOfTri(ax, ay, cx, cy, x, y);
	// Oh man...  there's some sort of small error
	// to account for.  Where does that come from?
	// And WHOA, running under QT on a mac, things are
	// MUCH less precise then they were under android.
	// What's the right amount of fudge to check for?
	return sum-0.01 < t;
}

float geo::degree(const float x2, const float y2) {
	if (y2 == 0) {
		if (x2 > 0) return 0.0;
		else if (x2 < 0) return 180.0;
		else return 0;
	}
	if (x2 == 0) {
		if (y2 > 0) return 90.0;
		else if (y2 < 0) return 270.0;
	}

	const float		x1 = 0, y1 = 0;
	const float		m1 = 0;
	const float		m2 = geo::slope(x1, y1, x2, y2);
	const float		tangent = (m2 - m1) / (1 + (m1 * m2));
	const float		radians = atan(tangent);
	const float		degree = radians * RTOD;

	if (x2 < 0 && y2 > 0) return degree + 180;
	else if (x2 < 0 && y2 < 0) return degree + 180;
	else if (x2 > 0 && y2 < 0) return degree + 360;

	return degree;
}

void geo::newPtOnLine(	const float x, const float y, const float  m, const float dist,
						Point3& newPt) {
	if (m == 0) {
		newPt.x = x + dist;
		newPt.y = y;
		return;
	} else if (!isfinite(m)) {
//	} else if (Double.isInfinite(m)) {
		newPt.x = x;
		newPt.y = y + dist;
		return;
	}

	/* This is basically a reverse engineering of the distance formula
	 * 		dist = sqrt(SQR(x2 - x1) + SQR(y2 - 1))
	 * It works by taking the final result -- the requested new distance,
	 * and deconstructing it into the xPart (SQR(x2 - x1) and the yPart,
	 * the sqrt of which can be added to the x and y values.
	 *
	 * The only trick is finding the xPart, which is just solved via
	 * the equation x + (x + (slope * slope)) = newDistance.
	 */
	const float		xPart = ((dist*dist) * (1 / (m * m))) / ( (1 / (m * m)) + 1);
	const float		yPart = xPart * (m * m);

	// Hmmm... just guessing...
	if (m < 0) {
		if (dist < 0) {
			newPt.x = x + sqrt(xPart);
			newPt.y = y - sqrt(yPart);
		} else {
			newPt.x = x - sqrt(xPart);
			newPt.y = y + sqrt(yPart);
		}
	} else {
		if (dist < 0) {
			newPt.x = x - sqrt(xPart);
			newPt.y = y - sqrt(yPart);
		} else {
			newPt.x = x + sqrt(xPart);
			newPt.y = y + sqrt(yPart);
		}
	}
}
