#include "raytri.h"

#include <math.h>
#include "math/geo.h"

using namespace intersection;

RayTri::RayTri()
{
}

int RayTri::on(	const Segment3& ray,
				const Point3& ptA, const Point3& ptB, const Point3& ptC,
				Point3& out)
{
	// get triangle edge vectors and plane normal
	mU.asSub(ptB, ptA);
	mV.asSub(ptC, ptA);
	mN.asCross3(mU, mV);
	// Don't deal with degenerate triangles
	if (mN.x == 0 && mN.y == 0 && mN.z == 0) return ERROR;

	mDir.asSub(ray.mPt[1], ray.mPt[0]);
	mW0.asSub(ray.mPt[0], ptA);
	const float			a = -mN.dot3(mW0),
						b = mN.dot3(mDir);
	if (fabs(b) < geo::EPS) {			// ray is parallel to triangle plane
		if (a == 0) return INSIDE;		// ray lies in triangle plane
		else return NO_INTERSECT;		// ray disjoint from plane
	}

	// get intersect point of ray with triangle plane
	const float			r = a / b;
	if (r < 0.0) return NO_INTERSECT;	// ray goes away from triangle
	// for a segment, also test if (r > 1.0) => no intersect
	mSegmentIntersects = (r < 1.0 + geo::EPS);

	// Intersect point of ray and plane
	out.set(ray.mPt[0].x + r * mDir.x,
			ray.mPt[0].y + r * mDir.y,
			ray.mPt[0].z + r * mDir.z);

	// is the intersection inside the triangle?
	const float			uu = mU.dot3(mU),
						uv = mU.dot3(mV),
						vv = mV.dot3(mV);
	mW.asSub(out, ptA);
	const float			wu = mW.dot3(mU),
						wv = mW.dot3(mV);
	const float			D = uv * uv - uu * vv;

	// get and test parametric coords
	const float			s = (uv * wv - vv * wu) / D;
	if (s < -geo::EPS || s > 1.0 + geo::EPS) return INTERSECT_PLANE;			// out is outside T
	const float			t = (uv * wu - uu * wv) / D;
	if (t < -geo::EPS || (s + t) > 1.0 + geo::EPS) return INTERSECT_PLANE;		// out is outside T

	return INTERSECT_TRI;
}

// Only use after performing the setup on a tri on the
// same plane as this tri.
int RayTri::onTri(	const Point3& a, const Point3& b, const Point3& c,
					Point3& out)
{
	mU.asSub(b, a);
	mV.asSub(c, a);

	const float			uu = mU.dot3(mU),
						uv = mU.dot3(mV),
						vv = mV.dot3(mV);
	mW.asSub(out, a);
	const float			wu = mW.dot3(mU),
						wv = mW.dot3(mV);
	const float			D = uv * uv - uu * vv;

	// get and test parametric coords
	const float			s = (uv * wv - vv * wu) / D;
	if (s < -geo::EPS || s > 1.0 + geo::EPS) return INTERSECT_PLANE;			// out is outside T
	const float			t = (uv * wu - uu * wv) / D;
	if (t < -geo::EPS || (s + t) > 1.0 + geo::EPS) return INTERSECT_PLANE;		// out is outside T

	return INTERSECT_TRI;
}
