#include "polygonclipper.h"

#include <QtDebug>
#include <QPointF>
#include <QPolygonF>
#include <math.h>

#include "math/geo.h"

PolygonClipper::PolygonClipper()
{
}

bool PolygonClipper::on(const QPolygonF& input, const QPolygonF& clip,
						QPolygonF& output)
{
	// special conditions
	if (input.size() < 1 || clip.size() < 1) {
		output.resize(0);
		return true;
	}

	mOutput.moveTo(mFactory);
	mInput.setTo(input, mFactory, true);
	mClip.setTo(clip, mFactory, false);
	if (mInput.mHead == NULL || mClip.mHead == NULL) return false;
	if (!makeIntersections(mInput, mClip)) {
		qDebug() << "ERROR PolygonClipper::on() can't make intersections";
		return false;
	}
	mInput.mHead->tail()->mHead = mInput.mHead;
	mClip.mHead->tail()->mHead = mClip.mHead;

	if (!startBuild(mInput, mClip, mStep)) {
		qDebug() << "ERROR PolygonClipper::on() can't start build";
		return false;
	}
//System.out.println("START BUILD on " + mStep.mVertex);

	const int			looping = mInput.size() + mClip.size() + 2;
	int					c = 0;

	mStep.mPrev = NULL;
	Vertex*				start = NULL;
	while (start != mStep.mVertex) {
		if (start == NULL) start = mStep.mVertex;

		Vertex*			tail = mFactory.pop(mStep.mVertex->mX, mStep.mVertex->mY);
		mOutput.addTail(tail);
//	System.out.println("ADDED " + mOutput.mHead.tail());
		if (!next(mStep)) break;
		mStep.mPrev = tail;

		if (c++ >= looping) {
			qDebug() << "ERROR PolygonClipper::on() cycle detected";
			return false;
		}
	}

//		System.out.println("RESULT");
//		mOutput.print();

	output.resize(0);
	Vertex*			v = mOutput.mHead;
	while (v != NULL) {
		output.append(QPointF(v->mX, v->mY));
		v = v->mNext;
	}
	return true;
}

bool PolygonClipper::makeIntersections(List& a, List& b)
{
	Vertex*				prvA = a.mHead->tail();
	Vertex*				curA = a.mHead;
	while (curA != NULL) {
		Vertex*			prvB = b.mHead->tail();
		Vertex*			curB = b.mHead;
		while (curB != NULL) {
			if (geo::lineSegmentIntersection(prvA->mX, prvA->mY, curA->mX, curA->mY, prvB->mX, prvB->mY, curB->mX, curB->mY, &mPt)) {
				Vertex*		iA = mFactory.pop(mPt.x, mPt.y);
				Vertex*		iB = mFactory.pop(mPt.x, mPt.y);
				if (iA == NULL || iB == NULL) {
					delete iA;
					delete iB;
					return false;
				}
				iA->mIntersection = iB;
				iB->mIntersection = iA;
				prvA->addWaitingIntersection(iA);
				prvB->addWaitingIntersection(iB);
			}
			prvB = curB;
			curB = curB->mNext;
		}

		prvA = curA;
		curA = curA->mNext;
	}

	// Place all intersections in the correct place in their segment
	a.insertWaiting();
	b.insertWaiting();

	// The intersection stage can produce a lot of identical points,
	// so remove them.  First weed out the non-intersections, since
	// intersections must always be preserved.
	a.removeExtranneousOriginals(mFactory);
	b.removeExtranneousOriginals(mFactory);
	a.removeExtranneousIntersections(b, mFactory);

	return true;
}

bool PolygonClipper::startBuild(List& a, List& b, Step& step) const
{
	Vertex*			v = a.mHead;
	while (v != NULL) {
		if (v->mIntersection == NULL && !(b.contains(v->mX, v->mY))) {
			step.mVertex = v;
			return true;
		}
		v = v->mNext;
	}
	return false;
}

bool PolygonClipper::next(Step& step) const
{
	if (!step.mVertex->isIntersection()) {
		step.mVertex = step.mVertex->circularNext();
		return true;
	}

	// OK, I'm an intersection.  I want to follow whatever point is
	// not in the clip list, or just continue, if both are.
	Vertex*			nextA = step.mVertex->circularNext();
	Vertex*			nextB = step.mVertex->mIntersection->circularNext();
	if (nextA == NULL || nextB == NULL) {
		qDebug() << "ERROR PolygonClipper::next(), no next";
		return false;
	}

	step.mVertex = turnRight(step, nextA, nextB);
	return true;
}

PolygonClipper::Vertex* PolygonClipper::turnRight(Step& step, Vertex* a, Vertex* b) const
{
	// Get the baseline degree
	const float			dir = geo::degree(step.mPrev->mX - step.mVertex->mX, step.mVertex->mY - step.mPrev->mY);
//System.out.println("\t" + step.mPrev + " to " + step.mVertex + " dir=" + dir);
	const float			aD = geo::degree(a->mX - step.mVertex->mX, step.mVertex->mY - a->mY),
						bD = geo::degree(b->mX - step.mVertex->mX, step.mVertex->mY - b->mY);
//System.out.println("\t\tad=" + aD + " is " + step.mVertex + " to " + a);
//System.out.println("\t\tbd=" + bD + " is " + step.mVertex + " to " + b);
	// If The baseline is at zero, then just continue CCW, which means
	// the highest number wins, with the exception of 0 itself
	if (dir == 0.0f) {
		if (aD == 0.0f) return a;
		if (bD == 0.0f) return b;
		if (aD >= bD) return a;
		return b;
	}
	// Now I have two cases to consider -- everything from myself down to
	// zero, and everything below that.
	const bool			aInTop = (aD <= dir && aD >= 0.0),
						bInTop = (bD <= dir && bD >= 0.0);
	if (aInTop && !bInTop) return a;
	if (!aInTop && bInTop) return b;
	// If they're both in the same half, larger wins.
	if (aD >= bD) return a;
	return b;
}
















// STEP
// --------------------------------------------
PolygonClipper::Step::Step()
	: mVertex(NULL)
	, mPrev(NULL)
{
}

// POLYGON-CLIPPER::VERTEX
// --------------------------------------------
PolygonClipper::Vertex::Vertex()
	: mNext(NULL)
	, mIntersection(NULL)
	, mHead(NULL)
	, mWaiting(NULL)
{
}

bool PolygonClipper::Vertex::isIntersection() const
{
	return mIntersection != NULL;
}

PolygonClipper::Vertex* PolygonClipper::Vertex::circularNext()
{
	if (mNext == NULL) return mHead;
	return mNext;
}

PolygonClipper::Vertex* PolygonClipper::Vertex::tail()
{
	Vertex*	o = this;
	while (o->mNext != NULL) o = o->mNext;
	return o;
}

void PolygonClipper::Vertex::addWaitingIntersection(Vertex* v)
{
	v->mWaiting = mWaiting;
	mWaiting = v;
}

void PolygonClipper::Vertex::insertWaiting()
{
	// Sort the list
	Vertex*			v = mWaiting;
	while (v != NULL) {
		v->mD = geo::distSquared(mX, mY, v->mX, v->mY);
		v = v->mWaiting;
	}
	while (sortWaiting()) ;
	// Insert everyone correctly
	v = mWaiting;
	while (v != NULL) {
		Vertex*		next = v->mWaiting;
		v->mNext = next;
		v->mWaiting = NULL;
		v = next;
	}
	mWaiting->tail()->mNext = mNext;
	mNext = mWaiting;
	mWaiting = NULL;
}

bool PolygonClipper::Vertex::sortWaiting()
{
	Vertex*				prev = NULL;
	Vertex*				cur = mWaiting;
	bool				changed = false;
	while (cur->mWaiting != NULL) {
		if (cur->mD > cur->mWaiting->mD) {
			changed = true;
			Vertex*		next = cur->mWaiting;
			if (prev == NULL) {
				cur->mWaiting = next->mWaiting;
				next->mWaiting = cur;
				mWaiting = next;
			} else {
				prev->mWaiting = next;
				cur->mWaiting = next->mWaiting;
				next->mWaiting = cur;
			}
			prev = next;
		} else {
			prev = cur;
			cur = cur->mWaiting;
		}
	}
	return changed;
}

// LIST
// --------------------------------------------
PolygonClipper::List::List()
	: mHead(NULL)
{
}

PolygonClipper::List::~List()
{
	Vertex*			v = mHead;
	while (v != NULL) {
		Vertex*		n = v->mNext;
		v->mNext = NULL;
		delete v;
		v = n;
	}
}

int PolygonClipper::List::size()
{
	int			c = 0;
	Vertex*		v = mHead;
	while (v != NULL) {
		c++;
		v = v->mNext;
	}
	return c;
}

void PolygonClipper::List::push(Vertex* v)
{
	if (v == NULL) {
//		System.out.println("ERROR asked to push a null vertex");
		return;
	}
	v->mNext = mHead;
	mHead = v;
}

PolygonClipper::Vertex* PolygonClipper::List::pop(const float x, const float y)
{
	Vertex*		ans = pop();
	if (ans == NULL) return NULL;
	ans->mX = x;
	ans->mY = y;
	return ans;
}

PolygonClipper::Vertex* PolygonClipper::List::pop()
{
	if (mHead == NULL) return new Vertex();
	Vertex*			ans = mHead;
	mHead = mHead->mNext;
	ans->mNext = NULL;
	ans->mIntersection = NULL;
	ans->mHead = NULL;
	ans->mWaiting = NULL;
	return ans;
}

void PolygonClipper::List::addTail(Vertex* v) {
	if (mHead == NULL) mHead = v;
	else mHead->tail()->mNext = v;
}

PolygonClipper::Vertex* PolygonClipper::List::remove(Vertex* v) {
	if (mHead == NULL) return NULL;
	if (v == NULL) {
//		System.out.println("ERROR asked to remove a null vertex");
		return NULL;
	}
	if (mHead == v) return pop();

	Vertex*		cur = mHead;
	while (cur->mNext != NULL) {
		if (cur->mNext == v) {
			cur->mNext = v->mNext;
			v->mNext = NULL;
			return v;
		}
		cur = cur->mNext;
	}
	return NULL;
}

void PolygonClipper::List::moveTo(List& o) {
	if (mHead != NULL) {
		if (o.mHead != NULL) mHead->tail()->mNext = o.mHead;
		o.mHead = mHead;
		mHead = NULL;
	}
}

bool PolygonClipper::List::contains(const float x, const float y) {
	Vertex*			prv = mHead->tail();
	Vertex*			cur = mHead;
	bool			ans = false;
	while (cur != NULL) {
		if (( ((cur->mY <= y) && (y < prv->mY))
				|| ((prv->mY <= y) && (y < cur->mY)))
				&& (x < (prv->mX - cur->mX) * (y - cur->mY) / (prv->mY - cur->mY) + cur->mX))
			ans = !ans;

		prv = cur;
		cur = cur->mNext;
	}
	return ans;
}

bool PolygonClipper::List::setTo(const QPolygonF& p, List& factory, const bool tails) {
	moveTo(factory);

	for (int k=0; k<p.size(); k++) {
		const QPointF&	pt = p.at(k);
		Vertex*			v = factory.pop(pt.x(), pt.y());
		if (v == NULL) return false;
		if (tails) addTail(v);
		else push(v);
	}
	return true;
}

void PolygonClipper::List::insertWaiting() {
	Vertex*		v = mHead;
	while (v != NULL) {
		if (v->mWaiting != NULL) v->insertWaiting();
		v = v->mNext;
	}
}

static const float		TOLERANCE = 0.000000001f;

static bool is_unique(	const float x1, const float y1,
						const float x2, const float y2) {
	return fabs(x1-x2) > TOLERANCE || fabs(y1-y2) > TOLERANCE;
}

bool PolygonClipper::List::isUnique(const Vertex& v1, const Vertex& v2) {
	return is_unique(v1.mX, v1.mY, v2.mX, v2.mY);
}

void PolygonClipper::List::removeExtranneousOriginals(List& factory) {
	while (mHead != NULL && mHead->mNext != NULL && !mHead->isIntersection()
			&& !isUnique(*mHead, *(mHead->mNext))) {
		factory.push(pop());
	}
	if (mHead == NULL) return;
	Vertex*				cur = mHead->mNext;
	while (cur->mNext != NULL) {
		if (!cur->mNext->isIntersection() && !isUnique(*cur, *(cur->mNext))) {
			Vertex*		pop = cur->mNext;
			cur->mNext = pop->mNext;
			pop->mNext = NULL;
			factory.push(pop);
		} else {
			cur = cur->mNext;
		}
	}
}

void PolygonClipper::List::removeExtranneousIntersections(List& b, List& factory) {
	while (mHead != NULL && mHead->mNext != NULL && mHead->isIntersection()
			&& mHead->mNext->isIntersection() && !isUnique(*mHead, *(mHead->mNext))) {
		factory.push(b.remove(mHead->mIntersection));
		factory.push(pop());
	}
	if (mHead == NULL) return;
	Vertex*				cur = mHead->mNext;
	while (cur->mNext != NULL) {
		if (cur->isIntersection() && cur->mNext->isIntersection() && !isUnique(*cur, *(cur->mNext))) {
			factory.push(b.remove(cur->mNext->mIntersection));
			Vertex*		pop = cur->mNext;
			cur->mNext = pop->mNext;
			pop->mNext = NULL;
			factory.push(pop);
		} else {
			cur = cur->mNext;
		}
	}
}
