#ifndef POLYGONCLIPPER_H
#define POLYGONCLIPPER_H

#include "math/primitives/point3.h"

class QPolygonF;

/* Clip one polygon by another.  Implemented by finding the intersection
 * points between polygons, removing duplicates (with a very high level of
 * precision), then applying a "turn right" rule to trace the minimal path.
 */
class PolygonClipper
{
public:
    PolygonClipper();

	// Subtract clip from input, placing the results in output.
	// If I return false then output is unchanged.
	bool on(const QPolygonF& input, const QPolygonF& clip,
			QPolygonF& output);

private:
	class Vertex;

	// STEP
	// --------------------------------------------
	class Step {
	public:
		Step();

		Vertex*			mVertex;
		Vertex*			mPrev;
	};

	// VERTEX
	// --------------------------------------------
	class Vertex {
	public:
		Vertex();

		Vertex*			mNext;
		float			mX, mY;
		Vertex*			mIntersection;	// If I have an intersection,
										// point to it in the other list
		Vertex*			mHead;
		Vertex*			mWaiting;		// While flattening the lists, I store
										// all intersections along this segment

		bool isIntersection() const;
		Vertex* circularNext();
		Vertex* tail();
		void addWaitingIntersection(Vertex* v);
		void insertWaiting();

	private:
		float			mD;				// only while sorting
		bool sortWaiting();

	};

	// LIST
	// --------------------------------------------
	class List {
	public:
		List();
		~List();

		Vertex*		mHead;

		int size();

		void push(Vertex* v);
		Vertex* pop(const float x, const float y);
		Vertex* pop();
		void addTail(Vertex* v);
		Vertex* remove(Vertex* v);
		void moveTo(List& o);

		bool contains(const float x, const float y);

		bool setTo(const QPolygonF& p, List& factory, const bool tails);

		void insertWaiting();

		// Remove all non-intersecting points that have identical intersections.
		void removeExtranneousOriginals(List& factory);
		// Remove all intersecting points that have duplicates.
		void removeExtranneousIntersections(List& b, List& factory);

	private:
		bool isUnique(const Vertex& v1, const Vertex& v2);
	};

	List				mInput, mClip, mOutput, mFactory;
	Step				mStep;
	Point3				mPt;

	bool				makeIntersections(List& a, List& b);
	// Find the first point that's in a but not b.
	bool				startBuild(List& a, List& b, Step& step) const;
	// Perform subtraction by examining all non-intersecting
	// vertices after the intersecting ones.  Choose the vertex
	// that is not in the subtraction list.
	bool				next(Step& step) const;
	// Choose the right-most vertex by looking at its degree in relation to
	// a start degree (which would be the previous and current vertices).
	// Since we build a clockwise list, right is always the degree closest
	// to the start, continuing counter clockwise.
	Vertex*				turnRight(Step& step, Vertex* a, Vertex* b) const;
};


#endif // POLYGONCLIPPER_H
