#include "quadclipper.h"

#include <math.h>
#include "math/geo.h"
#include "math/quadlist.h"

static bool might_intersect(const QuadList::Entry& e0, const QuadList::Entry& e1);

QuadClipper::QuadClipper()
{
#ifdef REPORT_GEO
	mBiAlgo.mReport = &mReport;
#endif
}

void QuadClipper::on(const QuadList& cmp, const float z, QuadList& out)
{
// qDebug() << "Clipper on z=" << z << " input size=" << cmp.size();
	// Step over each element of src, duplicating and
	// running against the src list.  Duplicated results
	// are pushed into the output.
	const QuadList::Entry*			src = cmp.head();
	if (src == NULL) return;
	// Keep track of my previous and next, which are assumed
	// to be my connected neighbors.  I don't clip against my neighbors.
	const QuadList::Entry*			prv = src->tail();

	QuadList						clipped;
//int		idx = 0;
	while (src != NULL) {
//qDebug() << "\t" << idx++;
		const QuadList::Entry*		nxt = src->mNext;
		if (nxt == NULL) nxt = cmp.head();

//		mResults.start(&(src->mQ));
		mSkyline.start(&(src->mQ));
		const QuadList::Entry*		dst = cmp.head();
		while (dst != NULL) {
//while (dst != NULL && src == cmp.head()) {
//			if (src != dst && src->overlaps(*dst)) {
//if (src != dst && dst != prv && dst != nxt) qDebug() << "overlap?";
//			if (src != dst && dst != prv && dst != nxt && src->overlaps(*dst)) {
			if (src != dst && might_intersect(*src, *dst)) {
				// Subtract the dst quad from the src
				if (mBiAlgo.bisect(src->mQ, dst->mQ, mBi)) {
//					mResults.slice(mBi);
					mSkyline.slice(mBi);
					mAccum.append(mBi);
#ifdef REPORT_GEO
					mReport.mSeg.append(mBi);
#endif
				}
			}
			dst = dst->mNext;
		}
		// Create new quads out of the subtraction results.
//		mResults.getResults(out, mPool);
//		mSkyline.getResults(out, mPool);
		if (!mSkyline.addTo(out, mPool)) {
			clipped.push(mPool.pop(src->mQ));
		}

		// step
		prv = src;
		src = src->mNext;
	}

	if (clipped.head() != NULL) {
		InnerShape			inner;
		inner.merge(out);
		inner.merge(mAccum);
		inner.getResults(clipped, out, mPool);
	}

	// Once all the quads are generated, scale the interior z to
	// match the requested.  Since a quad might be clipped at
	// the middle, we don't want to just set the interiors.
//	out.scaleToZ(z);
	QuadList::Entry*		e = out.head();
	while (e != NULL) {
		e->mQ.mPt[1].z = z;
		e->mQ.mPt[2].z = z;
		e = e->mNext;
	}
}

// RAY-INTERSECTS-SURFACE
// --------------------------------------------

// My data model lets quads collapse into tris on the inside.
static bool isInsideTri(const Quad& q) {
	return q.mPt[1].sameAs2(q.mPt[2]);
}

// My data model lets quads collapse into tris on the outside.
static bool isOutsideTri(const Quad& q) {
	return q.mPt[0].sameAs2(q.mPt[3]);
}

int QuadClipper::RayIntersectsSurface::on(const Segment3& ray, const Quad& q, Point3& out)
{
	int				ans = intersection::RayTri::ERROR;
	if (isInsideTri(q)) {
		ans = mRIT.on(ray, q.mPt[0], q.mPt[1], q.mPt[3], out);
		if (ans == intersection::RayTri::INTERSECT_TRI) return ans;
	} else if (isOutsideTri(q)) {
		ans = mRIT.on(ray, q.mPt[0], q.mPt[1], q.mPt[2], out);
		if (ans == intersection::RayTri::INTERSECT_TRI) return ans;
	} else {
		ans = mRIQ.on(ray, q, out);
	}
	return ans;
}

// INNER-SHAPE
// --------------------------------------------
void QuadClipper::InnerShape::merge(const QuadList& src)
{
	const QuadList::Entry*	e = src.head();
	while (e != NULL) {
		mPt.append(e->mQ.mPt[1]);
		mPt.append(e->mQ.mPt[2]);

		e = e->mNext;
	}
}

void QuadClipper::InnerShape::merge(const QList<Segment3>& src)
{
	float					maxZ = -1000000;
	for (int i=0; i<src.size(); i++) {
		const float			aZ = src[i].mPt[0].z,
							bZ = src[i].mPt[1].z;
		if (aZ > maxZ) maxZ = aZ;
		if (bZ > maxZ) maxZ = bZ;
	}
	maxZ -= 0.00001f;
	for (int i=0; i<src.size(); i++) {
		if (src[i].mPt[0].z > maxZ) mPt.append(src[i].mPt[0]);
		if (src[i].mPt[1].z > maxZ) mPt.append(src[i].mPt[1]);
//mPt.append(src[i].mPt[0]);
//mPt.append(src[i].mPt[1]);
	}
}

void QuadClipper::InnerShape::getResults(const QuadList& src, QuadList& dst, QuadList& pool) const
{
	Quad						q;
	const QuadList::Entry*		e = src.head();
	while (e != NULL) {
		const Point3*			p1 = nearest(e->mQ.mPt[0]);
		const Point3*			p2 = nearest(e->mQ.mPt[3]);
		if (p1 != NULL && p2 != NULL) {
			q.set(e->mQ.mPt[0], *p1, *p2, e->mQ.mPt[3]);
			dst.push(pool.pop(q));
		}
		e = e->mNext;
	}
}

const Point3* QuadClipper::InnerShape::nearest(const Point3& src) const
{
	if (mPt.size() < 1) return NULL;

	const Point3*		pt = &mPt[0];
	float				d = src.dist3(*pt);
	for (int i=1; i<mPt.size(); i++) {
		const Point3*	tpt = &mPt[i];
		const float		td = src.dist3(*tpt);
		if (td < d) {
			pt = tpt;
			d = td;
		}
	}
	return pt;
}

// SKYLINE
// --------------------------------------------
QuadClipper::Skyline::Skyline()
	: mList(NULL)
	, mAlloc(0)
	, mSize(0)
{
	Entry*		newList = (Entry*)realloc(mList, sizeof(Entry)*BLOCK_SIZE);
	if (newList != NULL) {
		mList = newList;
		mAlloc = BLOCK_SIZE;
	}
}

QuadClipper::Skyline::~Skyline()
{
	realloc(mList, 0);
}

void QuadClipper::Skyline::start(const Quad* q)
{
	mQ = q;
	mSize = 0;
}

void QuadClipper::Skyline::slice(const Segment3& s)
{
	add(s);
}

bool QuadClipper::Skyline::addTo(QuadList& out, QuadList& pool)
{
	if (mSize >= 1) return false;
	if (mQ->isInwardTri()) return false;

	out.push(pool.pop(*mQ));
	return true;
}

void QuadClipper::Skyline::getResults(QuadList& out, QuadList& pool)
{
	if (mSize < 1) {
		out.push(pool.pop(*mQ));
		return;
	} else qDebug() << "SKIP " << *mQ;
//out.push(pool.pop(*mQ));
}

bool QuadClipper::Skyline::add(const Segment3& s)
{
	if (mSize >= mAlloc && !growTo(mAlloc+BLOCK_SIZE)) return false;

	mList[mSize].mSeg.set(s);

	mSize++;
	return true;
}

bool QuadClipper::Skyline::growTo(const int newAlloc)
{
	if (newAlloc <= mAlloc) return true;
	Entry*		newList = (Entry*)realloc(mList, sizeof(Entry)*newAlloc);
	if (newList == NULL) return false;

	mList = newList;
	mAlloc = newAlloc;
	return true;
}

// RESULT
// --------------------------------------------
QuadClipper::Result::Result()
	: mQ(NULL)
	, mResS(0)
	, mResE(0)
{
}

QuadClipper::Result::~Result()
{
}

void QuadClipper::Result::start(const Quad* q)
{
	mQ = q;
	mPoly.resize(0);
}

void QuadClipper::Result::start()
{
	mPoly.append(QPointF(0, mQ->mPt[0].z));
	mPoly.append(QPointF(0, mQ->mPt[1].z));
	mPoly.append(QPointF(1, mQ->mPt[2].z));
	mPoly.append(QPointF(1, mQ->mPt[3].z));
}

// Given a quad b, find where it intersects my quad, and
// add the intersection (if any) to my list.
void QuadClipper::Result::slice(const Segment3& biIn)
{
//	qDebug() << "clip AT" << mBisector;
	float				startP = getPos(biIn.mPt[0], *mQ),
						endP = getPos(biIn.mPt[1], *mQ);
	// If the intersection is just along a line, bail
	if (startP == endP) return;

	Segment3			bi;
	bi.set(biIn);

	// If the slice is not in the same order as the quad, flip
	if (endP < startP) {
		const float		tp = startP;	startP = endP;	endP = tp;
		bi.reverse();
	}

	if (mPoly.size() < 1) start();

	const float			maxZ = fmax(mQ->mPt[1].z, mQ->mPt[2].z) + geo::EPS + 1;

	// Build a polygon out of the new slice, normalized to
	// my local coords.  Note that the subtracted polygon is
	// essentially flipped on the Y -- that is, it's low values
	// are the z's and its high values are infinite.
	mPolyB.resize(0);
	mPolyB.append(QPointF(startP, bi.mPt[0].z));
	mPolyB.append(QPointF(startP, maxZ));
	mPolyB.append(QPointF(endP, maxZ));
	mPolyB.append(QPointF(endP, bi.mPt[1].z));
	mClipper.on(mPoly, mPolyB, mPoly);
}

void QuadClipper::Result::getResults(QuadList& out, QuadList& pool)
{
	simplify(mPoly);

	if (mPoly.size() < 2 || !findResultIndexes()) {
		out.push(pool.pop(*mQ));
		return;
	}

	const Point3&			a = mQ->mPt[0];
	const Point3&			d = mQ->mPt[3];
	QPointF*				prev = &(mPoly[mResS]);
	mResS++;
	if (mResS >= mPoly.size()) mResS = 0;

	while (mResS != mResE) {
		QPointF*			next = &(mPoly[mResS]);
		// Generate a new quad
		QuadList::Entry*	e = pool.pop();
		e->mQ.mPt[0].set(at(prev->x(), getZ(prev->x(), a, d)));
		e->mQ.mPt[1].set(at(prev->x(), prev->y()));
		e->mQ.mPt[2].set(at(next->x(), next->y()));
		e->mQ.mPt[3].set(at(next->x(), getZ(next->x(), a, d)));
		out.push(e);

		prev = next;
		mResS++;
		if (mResS >= mPoly.size()) mResS = 0;
	}
}

void QuadClipper::Result::simplify(QPolygonF& p) const
{
	// To find the scaled pixel, take the largest distance and divide.
//	const float			d0 = geo::dist(mQ->mPt[0].x, mQ->mPt[0].y, mQ->mPt[3].x, mQ->mPt[3].y),
//						d1 = geo::dist(mQ->mPt[1].x, mQ->mPt[1].y, mQ->mPt[2].x, mQ->mPt[2].y);
//	const float			sPix = 1 / Math.max(d0, d1);
	const float			sPix = 0.1f;
	const float			minZ = fmax(mQ->mPt[0].z, mQ->mPt[3].z) + geo::EPS;
	// Combine everyone within range
/*
System.out.println("SIMPLIFY ON SCALED PIX=" + sPix);
System.out.println("\td0=" + d0 + " " + mQ.mPt[0] + " to " + mQ.mPt[3]);
System.out.println("\td1=" + d1 + " " + mQ.mPt[1] + " to " + mQ.mPt[2]);
System.out.println("BEFORE");
p.print();
*/
	for (int k=1; k<p.size(); k++) {
		QPointF&		prv = p[k-1];
		QPointF&		nxt = p[k];
		// If not the endpoints, or one on the bottom line...
//	System.out.println("\t" + k + " prv=" + prv + " nxt="  + nxt);
		if (prv.y() > minZ && nxt.y() > minZ) {
			if (fabs(prv.x()-nxt.x()) <= sPix) {
//				System.out.println("\t" + k + " simplify " + prv + " " + nxt);
				// Always maintain the edges
				if (prv.x() <= 0.5) {
					prv.setY(fmin(prv.y(), nxt.y()));
					p.remove(k);
					k--;
				} else {
					nxt.setY(fmin(prv.y(), nxt.y()));
					p.remove(k-1);
					k--;
				}
			}
		}
	}
//System.out.println("AFTER");
//p.print();
}

bool QuadClipper::Result::findResultIndexes()
{
	mResS = -1;
	mResE = -1;
	const float				MATCH_TOL = 0.0001;
	float					matchS = 10000, matchE = 10000, match;
	const float				startZ = mQ->mPt[0].z, endZ = mQ->mPt[3].z;
	for (int k=0; k<mPoly.size(); k++) {
		const QPointF&		pt = mPoly[k];
//		System.out.println("pt=" + pt + " matchS=" + resultQ(pt, startZ) + " matchE=" + resultQ(pt, endZ));
		if ((match=resultQ(pt, 0, startZ)) < MATCH_TOL && match < matchS) {
			mResS = k;
			matchS = match;
		} else if ((match=resultQ(pt, 1, endZ)) < MATCH_TOL && match < matchE) {
			mResE = k;
			matchE = match;
		}
	}
	// The value one past the start should be the start of the polyline.
	// If it's not at a pos of 0, something is wrong.
	if (mResS >= 0 && mResE >= 0) {
		mResS++;
		if (mResS >= mPoly.size()) mResS = 0;
//	if (mPoly.mPt[mResS].x > MATCH_TOL) System.out.println("ERR hheeeere x=" + mPoly.mPt[mResS].x);
//		if (mPoly.mPt[mResS].x > MATCH_TOL) return false;
		// Weird error
		if (mResS == mResE) return false;
		return true;
	}
	return false;
}

float QuadClipper::Result::resultQ(const QPointF& pt, const float x, const float y) const
{
	// Searching for 0, minZ
	return (fabs(pt.x()-x) + fabs(pt.y()-y)) / 2;
}

float QuadClipper::Result::getZ(const float pos, const Point3& a, const Point3& b) const
{
	return (a.z*(1-pos)) + (b.z*pos);
}

float QuadClipper::Result::getPos(const Point3& pt, const Quad& q) const
{
	const Point3&		a = q.mPt[0];
	const Point3&		b = q.mPt[1];
	const Point3&		c = q.mPt[2];
	const Point3&		d = q.mPt[3];
	// Find the points on either edge at my current z.  Assumes
	// a.z is < b.z and d.z < c.z.
	const float			edgeAb = (pt.z - a.z) / (b.z - a.z),
						edgeCd = (pt.z - d.z) / (c.z - d.z);
	const float			abX = (a.x*(1-edgeAb)) + (b.x*edgeAb),
						abY = (a.y*(1-edgeAb)) + (b.y*edgeAb),
						cdX = (d.x*(1-edgeCd)) + (c.x*edgeCd),
						cdY = (d.y*(1-edgeCd)) + (c.y*edgeCd);
	// Find the distance between my point and its closest
	// edge points.
	const float			d0 = geo::dist(pt.x, pt.y, abX, abY),
						d1 = geo::dist(cdX, cdY, pt.x, pt.y);
	return d0 / (d0 + d1);
}

Point3& QuadClipper::Result::at(const float pos, const float z)
{
	const Point3&		a = mQ->mPt[0];
	const Point3&		b = mQ->mPt[1];
	const Point3&		c = mQ->mPt[2];
	const Point3&		d = mQ->mPt[3];
	// Find the points on either edge at my current z.  Assumes
	// a.z is < b.z and d.z < c.z.
	const float			edgeAb = (z - a.z) / (b.z - a.z),
						edgeCd = (z - d.z) / (c.z - d.z);
	const float			abX = (a.x*(1-edgeAb)) + (b.x*edgeAb),
						abY = (a.y*(1-edgeAb)) + (b.y*edgeAb),
						cdX = (d.x*(1-edgeCd)) + (c.x*edgeCd),
						cdY = (d.y*(1-edgeCd)) + (c.y*edgeCd);
	mPt.set((abX*(1-pos)) + (cdX*pos), (abY*(1-pos)) + (cdY*pos), z);
	return mPt;
}

// QIQ-ALGO
// --------------------------------------------
QuadClipper::QiqAlgo::QiqAlgo()
{
}

bool QuadClipper::QiqAlgo::bisect(	const Quad& a, const Quad& b,
									Segment3& bisection)
{
#ifdef REPORT_GEO
	mQIQ.mReport = mReport;
#endif
	return mQIQ.on(a, b, bisection) > intersection::QuadQuad::ERROR;
}

// RAY-ALGO
// --------------------------------------------
QuadClipper::RayAlgo::RayAlgo()
{
}

bool QuadClipper::RayAlgo::bisect(	const Quad& a, const Quad& b,
									Segment3& bisector)
{
	// Shoot rays through all my cases until I find the intersections.
	// Need to be smart about the order in which I do this, so duplicates
	// don't mess the results up.
	mBiIdx = 0;

	find(a.mPt[0], a.mPt[1], b, 0, bisector);
	if (find(a.mPt[3], a.mPt[2], b, 1, bisector)) return true;
	// Cases where only a corner of a juts into b
	if (mF[0] == intersection::RayTri::INTERSECT_TRI) {
		if (find(a.mPt[1], a.mPt[2], b, 4, bisector)) return true;
	}
	if (mF[1] == intersection::RayTri::INTERSECT_TRI) {
		if (find(a.mPt[2], a.mPt[1], b, 4, bisector)) return true;
	}
	// Cases where the edges of b intersect a
	if (find(b.mPt[0], b.mPt[1], a, 2, bisector)) return true;
	if (find(b.mPt[3], b.mPt[2], a, 3, bisector)) return true;
	// Cases where only a corner of b juts into a
	if (mF[2] == intersection::RayTri::INTERSECT_TRI) {
		if (find(b.mPt[1], b.mPt[2], a, 4, bisector)) return true;
	}
	if (mF[3] == intersection::RayTri::INTERSECT_TRI) {
		if (find(b.mPt[2], b.mPt[1], a, 4, bisector)) return true;
	}
	return false;
}

bool QuadClipper::RayAlgo::find(	const Point3& seg0, const Point3& seg1,
									const Quad& q, const int idx, Segment3& bisector)
{
	mRay.mPt[0].set(seg0);
	mRay.mPt[1].set(seg1);
	mF[idx] = mRIS.on(mRay, q, mPt);
	if (mF[idx] < intersection::RayTri::INTERSECT_PLANE) mF[idx] = 0;
	if (mF[idx] == intersection::RayTri::INTERSECT_TRI) {
		if (mBiIdx < 2) pushIntersection(mPt, bisector);
	}
	return mBiIdx > 1;
}

void QuadClipper::RayAlgo::pushIntersection(const Point3& pt, Segment3& bisector)
{
	// Reject overlapping points.  I need distinct ones!
	for (int k=0; k<mBiIdx; k++) {
		if (bisector.mPt[k].sameAs2(pt)) return;
	}
	bisector.mPt[mBiIdx++].set(pt);
}

// misc
// --------------------------------------------
static bool might_intersect(const QuadList::Entry& e0, const QuadList::Entry& e1)
{
	if (!e0.mFrame.intersects((e1.mFrame))) return false;
	// Disable intersection for adjacent quads, determined by the
	// outer points being connected.
	if (e0.mQ.mPt[0].sameAs2(e1.mQ.mPt[3])) return false;
	if (e0.mQ.mPt[3].sameAs2(e1.mQ.mPt[0])) return false;
	return true;
}

