#ifndef QUADCLIPPER_H
#define QUADCLIPPER_H

#include <QPolygonF>
#include "math/polygonclipper.h"
#include "math/quadlist.h"
#include "math/intersection/rayquad.h"
#include "math/intersection/quadquad.h"
#include "math/intersection/segseg.h"
#include "math/report.h"

// Turning this on also turns on a new quad clipping routine
// that SHOULD be much better but is not at all.
//#define STORE_BISECTORS (1)
#define TEST_PIP (1)

// Really need to replace this with plane/line intersection
// http://softsurfer.com/Archive/algorithm_0104/algorithm_0104B.htm#intersect3D_2Planes()

/* Used to do subtraction on quads in 3d space.
 * Any quads that intersect on the Z plane will
 * be clipped at the line of intersection.
 ********************************************/
class QuadClipper
{
public:
    QuadClipper();

	// Find all intersections, chop them at the dividing line.
	// The chop point gets assigned the Z value.
	void on(const QuadList& cmp, const float z, QuadList& out);

#ifdef REPORT_GEO
	Report				mReport;
#endif

	QList<Segment3>		mAccum;

private:
	// RAY-INTERSECTS-SURFACE.  A special intersection test,
	// since I've chosen to operate on a quad-based format
	// that allows degenerate quads (forming triangles).
	// --------------------------------------------
	class RayIntersectsSurface {
	public:
		int on(const Segment3& ray, const Quad& q, Point3& out);

	private:
		intersection::RayQuad		mRIQ;
		intersection::RayTri		mRIT;
	};

	// BISECTOR-ALGO.  Interface for classes that
	// implement bisector-finding.
	// --------------------------------------------
	class BisectorAlgo {
	public:
		virtual ~BisectorAlgo()	{ }

		virtual bool	bisect(	const Quad& a, const Quad& b,
								Segment3& bisection) = 0;

	protected:
		BisectorAlgo()			{ }

#ifdef REPORT_GEO
	public:
		Report			*mReport;
#endif
	};

	// INNER-SHAPE.  Store the inner points.  Inner
	// points are either the unclipped values of
	// a non-degenerate quad, or the highest Z pts
	// from the intersections.
	// --------------------------------------------
	class InnerShape {
	public:
		// Add all the inner points to me
		void					merge(const QuadList&);
		// Add the highest points to me
		void					merge(const QList<Segment3>&);

		void					getResults(const QuadList&, QuadList& dst, QuadList& pool) const;

	private:
		QList<Point3>			mPt;

		const Point3*			nearest(const Point3& src) const;
	};

	// SKYLINE.  The workhorse of the algo, store
	// the intersection line for each intersecting
	// quad.
	// --------------------------------------------
	class Skyline {
	public:
		Skyline();
		~Skyline();

		void						start(const Quad* q);
		void						slice(const Segment3& bi);
		// If I have no intersections and am valid, add me.
		// Pool is a global resource for recycling quad entries.
		// Note that my original quad is immutable.
		bool						addTo(QuadList& out, QuadList& pool);

		// Generate results based on my list of intersections.
		// Pool is a global resource for recycling quad entries.
		// Note that my original quad is immutable.
		void						getResults(QuadList& out, QuadList& pool);

		private:
			const Quad*				mQ;

			Segment3				mBisector;
			intersection::QuadQuad	mQIQ;
			intersection::SegSeg	mSS;


			bool					add(const Segment3&);
			bool					growTo(const int newAlloc);

			static const int		BLOCK_SIZE = 16;

			class Entry {
			public:
				float				mStart, mStop;
				Segment3			mSeg;
			};

			Entry*					mList;
			int						mAlloc, mSize;
	};
	Skyline							mSkyline;

	// RESULT.  Store a polygon 2d projection of the
	// quad being clipped, and subtract from it
	// each clipping polygon.
	// --------------------------------------------
	class Result {
	public:
		Result();
		~Result();

		void start(const Quad* q);

		// Given a quad b, find where it intersects my quad, and
		// add the intersection (if any) to my list.
		void slice(const Segment3& bi);

		// Generate results based on my slice list and place them on the out.
		// Pool is a global resource for recycling quad entries.  Note that
		// my original entry is immutable.
		void getResults(QuadList& out, QuadList& pool);

	private:
		const Quad*					mQ;

		QPolygonF					mPoly, mPolyB;
		PolygonClipper				mClipper;

		// When building the results, these are the start and end indexes
		int							mResS, mResE;

		Point3						mPt;

		void start();

		// The polygon subtraction operation runs with a very high degree
		// of precision, but that means the results can be too much geometry
		// for the display.  Do this by looking at (roughly) the scaled value
		// of a single pixel, then combining pts within range.
		void simplify(QPolygonF&) const;

		// Find the result indexes that will iterate over the polygon
		// as if it's a polyline, ignoring the bottom left and right corners,
		// only looking at the high z values.  Every result polygon will have
		// a (0, minZ) and a (1, minZ) point, so find them.
		bool findResultIndexes();

		float resultQ(const QPointF& pt, const float x, const float y) const;
		float getZ(const float pos, const Point3& a, const Point3& b) const;

		// Answer the normalized location of pt between the edges of a
		float getPos(const Point3& pt, const Quad& q) const;
		// Given a normalized location, find its x,y
		Point3& at(const float pos, const float z);
	};

	QuadList			mPool;
	Result				mResults;

	// QIQ-ALGO.  Find bisection with quad-intersection.
	// --------------------------------------------
	class QiqAlgo : public BisectorAlgo {
	public:
		QiqAlgo();
		virtual bool	bisect(	const Quad& a, const Quad& b,
								Segment3& bisection);

	private:
		intersection::QuadQuad	mQIQ;
	};

	// RAY-ALGO.  Find bisection by sending out rays.
	// --------------------------------------------
	class RayAlgo : public BisectorAlgo {
	public:
		RayAlgo();
		virtual bool			bisect(	const Quad& a, const Quad& b,
										Segment3& bisection);

	private:
		int						mBiIdx;
		int						mF[5];
		Point3					mPt;
		Segment3				mRay;
		RayIntersectsSurface	mRIS;

		// Test the seg against both triangles for the target quad.
		// Return true if I've found my line of intersection.
		bool					find(	const Point3& seg0, const Point3& seg1,
										const Quad& q, const int idx, Segment3& bisector);
		void					pushIntersection(const Point3& pt, Segment3& bisector);
	};

	QiqAlgo				mBiAlgo;
//	RayAlgo				mBiAlgo;
	Segment3			mBi;
};

#endif // QUADCLIPPER_H
