#include "quadlist.h"

#include <math.h>

/* QUAD-LIST
 ********************************************/
QuadList::QuadList()
	: mHead(NULL)
{
}

QuadList::~QuadList()
{
	deleteList();
}

int QuadList::size() const
{
	const Entry*	e = mHead;
	int				size = 0;
	while (e != NULL) {
		size++;
		e = e->mNext;
	}
	return size;
}

QuadList::Entry* QuadList::head()
{
	return mHead;
}

const QuadList::Entry* QuadList::head() const
{
	return mHead;
}

void QuadList::push(Entry* e)
{
	if (e == NULL) return;
	e->mNext = mHead;
	mHead = e;
}

bool QuadList::push(const Point3& a, const Point3& b,
					const Point3& c, const Point3& d)
{
	Entry*		e = new Entry();
	if (e == NULL) return false;
	e->mQ.set(a, b, c, d);
	e->cacheFrame();
	push(e);
	return true;
}

void QuadList::addTail(Entry* e)
{
	if (e == NULL) return;
	if (mHead == NULL) mHead = e;
	else mHead->tail()->mNext = e;
}

QuadList::Entry* QuadList::pop()
{
	if (mHead == NULL) return new Entry();
	Entry*		e = mHead;
	mHead = mHead->mNext;
	e->mNext = NULL;
	return e;
}

QuadList::Entry* QuadList::pop(const Quad& q)
{
	Entry*		e = pop();
	e->mQ.set(q);
	e->cacheFrame();
	return e;
}

QuadList::Entry* QuadList::detach()
{
	Entry*		ans = mHead;
	mHead = NULL;
	return ans;
}

void QuadList::attach(Entry* e)
{
	deleteList();
	mHead = e;
}

void QuadList::swap(QuadList& o)
{
	Entry*		e = mHead;
	mHead = o.mHead;
	o.mHead = e;
}

void QuadList::	scaleToZ(const float z)
{
	if (mHead == NULL) return;
	Entry*			dst = mHead;

	float			max = dst->mQ.mPt[1].z;
	while (dst != NULL) {
		if (dst->mQ.mPt[1].z > z) max = dst->mQ.mPt[1].z;
		if (dst->mQ.mPt[2].z > z) max = dst->mQ.mPt[2].z;
		dst = dst->mNext;
	}
	if (z == 0 || max == 0) return;
	const float		scale = z / max;
	dst = mHead;
	while (dst != NULL) {
		dst->mQ.mPt[1].z *= scale;
		dst->mQ.mPt[2].z *= scale;
		dst = dst->mNext;
	}
}

void QuadList::deleteList()
{
	Entry*		e = mHead;
	while (e != NULL) {
		Entry*	n = e->mNext;
		e->mNext = NULL;
		delete e;
		e = n;
	}
	mHead = NULL;
}

/* QUAD-LIST::ENTRY
 ********************************************/
QuadList::Entry::Entry()
	: mNext(NULL)
{
}

QuadList::Entry* QuadList::Entry::tail()
{
	Entry*		e = this;
	while (e->mNext != NULL) e = e->mNext;
	return e;
}

const QuadList::Entry* QuadList::Entry::tail() const
{
	const Entry*		e = this;
	while (e->mNext != NULL) e = e->mNext;
	return e;
}

bool QuadList::Entry::overlaps(const Entry& o) const
{
	return mFrame.intersects(o.mFrame);
}

void QuadList::Entry::cacheFrame()
{
	mFrame.setLeft(mQ.mPt[0].x);
	mFrame.setRight(mQ.mPt[0].x);
	mFrame.setTop(mQ.mPt[0].y);
	mFrame.setBottom(mQ.mPt[0].y);
	for (int k=1; k<Quad::SIZE; k++) {
		mFrame.setLeft(fmin(mFrame.left(), mQ.mPt[k].x));
		mFrame.setTop(fmin(mFrame.top(), mQ.mPt[k].y));
		mFrame.setRight(fmax(mFrame.right(), mQ.mPt[k].x));
		mFrame.setBottom(fmax(mFrame.bottom(), mQ.mPt[k].y));
	}
}

QDebug& operator<<(QDebug& d, const QuadList& o)
{
	d << "QuadList size=" << o.size();
	const QuadList::Entry*		e = o.head();
	while (e != NULL) {
		d << "\n\t" << e->mQ;
		e = e->mNext;
	}
	return d;
}
