#include "activeproject.h"

#include <QtDebug>
#include <QMutexLocker>

// There are no layers right now, so skip to the root.
static const int	INIT_DEPTH = fx::VisitParams::ROOT_DEPTH-1;

/* ACTIVE-PROJECT
 ******************************************************************/
ActiveProject::ActiveProject(	const ArcList& arcs, const ColorSrcList& cs,
								QObject *parent)
	: QObject(parent)
	, mFxUniqueId(0)
	, mArcs(arcs)
	, mClrSrc(cs)
	, mTimeStamp(0)
{
	mPrj.mTimeStamp = mTimeStamp;
}

ActiveProject::~ActiveProject()
{
}

int ActiveProject::nextFxUniqueId()
{
	return mFxUniqueId++;
}

bool ActiveProject::getArc(ArcHead& head, const int idx) const
{
	return mArcs.getArc(idx, head);
}

bool ActiveProject::getColorSrc(ColorSrcHead& h, const ColorSrcMatch& m) const
{
	return mClrSrc.find(h, m);
}

void ActiveProject::setFx(Fx* fx)
{
	{
		QMutexLocker		lock(&mLock);
		mPrj.mHead.setFx(fx);
	}
	emit projectChanged(STRUCTURE_CHANGED);
}

bool ActiveProject::copyProject(Project& dst) const
{
	QMutexLocker		lock(&mLock);
	return dst.copyFrom(mPrj);
}

bool ActiveProject::contains(const param::Url& url) const
{
	QMutexLocker			lock(&mLock);
	return mPrj.mHead.findParent(url) != NULL;
}

void ActiveProject::forEach(fx::Visitor& v)
{
	fx::VisitParams			vp;
	vp.mDepth = INIT_DEPTH;

	QMutexLocker			lock(&mLock);
	mPrj.mHead.forEach(v, vp);
}

void ActiveProject::getParams(param::List& list, param::GetArgs& get) const
{
	get.mError = 0;
	QMutexLocker		lock(&mLock);
	if (get.mUrl.size() < 1) {
		mPrj.mHead.getParams(list, get);
	} else {
		// If users supply a URL, then track down the parent and limit
		// the params to that.  If no parent, then do nothing (as opposed
		// to, say, defaulting to getting every param in the project,
		// which would be odd).
		const param::Parent*	p = mPrj.mHead.findParent(get.mUrl);
		if (p != NULL) {
			p->getParams(list, get);
			list.prepend(get.mUrl);
		}
	}
}

void ActiveProject::setParamF(const param::Url& url, const float v)
{
	mSetParam.clear();
	mSetParam.mV = v;
	setParam(url, mSetParam, false);
}

void ActiveProject::setParamStr(const param::Url& url, const QString& v)
{
	mSetParam.clear();
	mSetParam.mStr = v;
	setParam(url, mSetParam, false);
}

void ActiveProject::setParamArc(const param::Url& url, const ArcHead& h)
{
	mSetParam.clear();
	mSetParam.mArc.setTo(h);
	setParam(url, mSetParam, true);
}

void ActiveProject::setParamClr(const param::Url& url, const ColorSrcHead& h)
{
	mSetParam.clear();
	mSetParam.mClr = h;
	setParam(url, mSetParam, true);
}

void ActiveProject::setParam(const param::Url& url, param::SetArgs& sp, const bool structure)
{
	sp.mChanged = false;
	param::UrlIterator		it(url);
	if (!it.hasValue()) return;
	{
		// XXX Ideally, clients have a token that caches the
		// found parent, so I don't have to look for it every time.
		// Currently the expense is trivial, and for this app, that
		// will always be the case, but if the param framework is
		// ever used in a realtime environment, it would be nice.
		QMutexLocker		lock(&mLock);
		Fx*					fx = mPrj.mHead.fx();
		// Locate the param.  Need to match at the top level, or else DEATH.
		if (fx == NULL || !it.matches(fx->getId())) return;
		++it;
		param::Parent*		p = fx->findParent(it);
		if (p == NULL) return;
		p->setParam(it, sp);
	}
	if (sp.mChanged) {
		if (structure) emit projectChanged(STRUCTURE_CHANGED);
		else {
			Param			p;
			p.mUrl = url;
			sp.writeTo(p);
			emit paramChanged(p);
		}
	}
}
