#include "shapebrowser.h"

#include <QtGui>
#include "app/defs.h"
#include "io/svgreader.h"
#include "model/project.h"

static bool transform_frame(	const QRectF& srcF, const int dstS, QTransform& trans,
								const Look&);

ShapeBrowser::ShapeBrowser(	const Environment& env, Project& prj,
							const Look& look, QWidget *parent)
	: QWidget(parent)
	, mEnv(env)
	, mPrj(prj)
	, mLook(look)
	, mHead(NULL)
	, mLayout(true)
{
	mEntryH = mLook.mFh * 2 + mLook.mPadY * 3;
	mViewH = mEntryH + mLook.mPadY * 2;
	setMaximumHeight(mViewH);

	const QDir			dir(mEnv.getInstallPath(SHAPES_DIR));
	const QString		pdir(dir.absolutePath());
	const QStringList	files = dir.entryList(QDir::Files);
	for (int k=0; k<files.size(); k++) {
		add(pdir + QDir::separator() + files.at(k));
	}
}

QSize ShapeBrowser::minimumSizeHint() const
{
	return QSize(-1, mViewH);
}

void ShapeBrowser::mouseReleaseEvent(QMouseEvent* me)
{
	if (mHead == NULL) return;

	const Entry*		e = mHead->find(me->x(), me->y());
	if (e != NULL && e->mFilename.size() > 0) {
		mPrj.setShape(0, e->mFilename);
	}
}

void ShapeBrowser::paintEvent(QPaintEvent * /* event */)
{
	if (mLayout) layoutEntries();

	QPainter painter(this);
	// Dashboard background
	painter.setPen(mLook.mDashBgC);
	painter.setBrush(mLook.mDashBgB);
	painter.drawRect(0, 0, width(), height());

	Entry*		e = mHead;
	while (e != NULL) {
		painter.drawPixmap(e->mFrame.left(), e->mFrame.top(), e->mImg);
		e = e->mNext;
	}
}

bool ShapeBrowser::add(const QString& filename)
{
//	qDebug() << "add entry" << filename;

	Entry*		e = new Entry();
	if (e == NULL || !e->setImg(filename, mEntryH, mLook)) {
		delete e;
		return false;
	}

	e->mFilename = filename;

	if (mHead == NULL) mHead = e;
	else mHead->tail()->mNext = e;
	mLayout = true;
	return true;
}

void ShapeBrowser::layoutEntries()
{
	mLayout = false;

	Entry*		e = mHead;
	int			x = mLook.mPadX, top = mLook.mPadY;
	while (e != NULL) {
		e->mFrame.setRect(x, top, mEntryH, mEntryH);

		x = e->mFrame.right() + mLook.mPadX;
		e = e->mNext;
	}
}

/* ENTRY class
 *************************************/
ShapeBrowser::Entry::Entry()
	: mNext(NULL)
{
}

ShapeBrowser::Entry* ShapeBrowser::Entry::tail()
{
	Entry*		e = this;
	while (e->mNext != NULL) e = e->mNext;
	return e;
}

const ShapeBrowser::Entry* ShapeBrowser::Entry::find(const int x, const int y) const
{
	const Entry*		e = this;
	while (e != NULL) {
		if (e->mFrame.contains(x, y)) return e;
		e = e->mNext;
	}
	return NULL;
}

bool ShapeBrowser::Entry::setImg(const QString& filename, const int size, const Look& look)
{
	// Load a path for the underlying SVG file.
	SvgReader		reader;
	QPainterPath	path;
	if (!reader.load(filename, &path)) {
		qDebug() << "ERROR LOADING " << filename;
	}

	QImage*			img = new QImage(size, size, QImage::Format_RGB888);
	if (img == NULL) return false;

	{
		QPainter		painter(img);
		// Clear the background
		painter.setPen(Qt::NoPen);
		painter.setBrush(look.mDashBgB); //.mAlphaB);
		painter.drawRect(0, 0, size, size);
		// Draw a param panel
		painter.setBrush(look.mParamBgB);
		painter.setRenderHint(QPainter::Antialiasing);
		painter.drawRoundedRect(0, 0, size, size, look.mRound, look.mRound);
		// Translate the path to local coords
		const QRectF	pathF(path.boundingRect());
		QTransform		trans;
		transform_frame(pathF, size, trans, look);
		painter.setTransform(trans);
		// Draw the path
		painter.setBrush(look.mParamFgB);
		painter.drawPath(path);
	}

	// Kinda weird -- where's the error check?
	mImg = QPixmap::fromImage(*img);
	delete img;
	return true;
}

/* MISC
 *******************************/
static bool transform_frame(const QRectF& srcF, const int dstS, QTransform& trans, const Look& look)
{
	QPolygonF		a, b;
	const int		w = dstS - look.mPadX*2, h = dstS - look.mPadY*2;
	QRectF			dstF(look.mPadX, look.mPadY, w-1, h-1);
	a << srcF.topLeft() << srcF.topRight() << srcF.bottomRight() << srcF.bottomLeft();
	b << dstF.topLeft() << dstF.topRight() << dstF.bottomRight() << dstF.bottomLeft();

	return QTransform::quadToQuad(a, b, trans);
}
