#include "stripview.h"

#include <QDebug>
#include <QMouseEvent>
#include <QPainter>

#include "param/param.h"
#include "ui/look.h"

#include "ui/browserpopup.h"

/* STRIP-VIEW
 *******************************/
StripView::StripView(ActiveProject& prj, const ArcList& arcs,
					 const Look& look, QWidget *parent)
	: QWidget(parent)
	, mPrj(prj)
	, mArcs(arcs)
	, mLook(look)
	, mLayout(true)
	, mPlug(NULL)
{
	mEntryH = mLook.mFh * 2 + mLook.mPadY * 3;
	mViewH = mEntryH + mLook.mPadY * 2;
	setMaximumHeight(mViewH);
}

StripView::~StripView()
{
	setPlug(NULL);
}

QSize StripView::minimumSizeHint() const
{
	return QSize(-1, mViewH);
}

int StripView::entryH() const
{
	return mEntryH;
}

void StripView::setPlug(FxUi* p)
{
	// Clean up any entries
	mList.clear();

	// Replace the plug
	delete mPlug;
	mPlug = p;
	if (mPlug != NULL) mPlug->setParent(this, mLook);

	mLayout = true;
	update();
}

void StripView::add(FxUi::Entry* e)
{
	if (e == NULL) return;

	mList.append(e);
	mLayout = true;
	update();
}

void StripView::addBrowser(fxui::BrowserModel* model)
{
	if (model == NULL) return;

	FxUi::Entry*	tmp = new BrowserPopUp(*this, *this, model);
	if (tmp != NULL) {
		mList.append(tmp);
		mLayout = true;
		update();
	}
}

void StripView::onProjectChanged(const int changedFlags)
{
	if (mPlug == NULL) return;

	const int			ans = mPlug->onProjectChanged();
	if (ans == FxUi::UI_UNCHANGED) return;

	if (ans == FxUi::UI_ERROR) setPlug(NULL);
	else {
		mList.clear();
		mPlug->setParent(this, mLook);
	}

	mLayout = true;
	update();
}

void StripView::onParamChanged(const Param& p)
{
	for (int i=0; i<mList.size(); i++) {
		Holder&			h = mList[i];
		if (h.mEntry == NULL) continue;

		const int		ans = h.mEntry->onParamChanged(p);
		if (ans == FxUi::Entry::NO_PARAMS) return;
		if (ans == FxUi::Entry::PARAM) {
			update();
			return;
		}
	}
}

void StripView::mousePressEvent(QMouseEvent* me)
{
	mDown.mEntry = NULL;
	const Holder*		h = find(me->x(), me->y());
	if (h == NULL || h->mEntry == NULL) return;

	mDown.mEvent = me;
	mDown.mEntry = h->mEntry;
	mDown.mFrame = h->mFrame;
	// Special arc browser.
//	if (h->mEntry->paramType() == Param::ARC_BROWSER_TYPE) {
//		mOverlay = &mArcList;
//		update();
//	} else {
		mDown.mEntry->mousePressEvent(mDown);
//	}
	mDown.mEvent = NULL;
}

void StripView::mouseMoveEvent(QMouseEvent* me)
{
	if (mDown.mEntry != NULL) {
		mDown.mEvent = me;
		mDown.mEntry->mouseMoveEvent(mDown);
		mDown.mEvent = NULL;
	}
}

void StripView::mouseReleaseEvent(QMouseEvent* me)
{
	if (mDown.mEntry == NULL) return;

	FxUi::Entry*		entry = mDown.mEntry;
	const Holder*		h = find(me->x(), me->y());
	if (h == NULL || h->mEntry == NULL || h->mEntry != mDown.mEntry) mDown.mEntry = NULL;

	mDown.mEvent = me;
	entry->mouseReleaseEvent(mDown);
	mDown.mEvent = NULL;
	mDown.mEntry = NULL;
}

void StripView::paintEvent(QPaintEvent*)
{
	if (mLayout) layoutEntries();

	QPainter			painter(this);
	// Dashboard background
	painter.setPen(mLook.mDashBgC);
	painter.setBrush(mLook.mDashBgB);
	painter.drawRect(0, 0, width(), height());

	mPaintArgs.mPainter = &painter;
	mPaintArgs.mLook = &mLook;
	paintEntries(false);
	paintEntries(true);
}

void StripView::paintEntries(const bool drawOnTop)
{
	for (int i=0; i<mList.size(); i++) {
		Holder&			h = mList[i];
		if (h.mEntry == NULL) continue;

		const bool		isOnTop = (h.mEntry->flags()&FxUi::Entry::DRAW_ON_TOP_F) != 0;
		if (drawOnTop == isOnTop) {
			mPaintArgs.mFrame = h.mFrame;
			h.mEntry->onPaint(mPaintArgs);
		}
	}
}

void StripView::layoutEntries()
{
	mLayout = false;

	int					x = mLook.mPadX, top = mLook.mPadY;
	for (int i=0; i<mList.size(); i++) {
		Holder&			h = mList[i];
		if (h.mEntry == NULL) continue;

		const QSize		size = h.mEntry->size();
		h.mFrame.setRect(x, top, size.width(), size.height());
		x = h.mFrame.right();
		if ((h.mEntry->flags()&FxUi::Entry::CONNECT_RIGHT) == 0)
			x += mLook.mPadX;
	}
}

const StripView::Holder* StripView::find(const int x, const int y) const
{
	for (int i=0; i<mList.size(); i++) {
		if (mList[i].mFrame.contains(x, y)) return &mList[i];
	}
	return NULL;
}

/* HOLDER
 *******************************/
StripView::Holder::Holder()
	: mEntry(NULL)
{
}

StripView::Holder::Holder(FxUi::Entry* e)
	: mEntry(e)
{
}

StripView::Holder::Holder(const Holder& o)
	: mFrame(o.mFrame)
	, mEntry(o.mEntry)
{
}

void StripView::Holder::operator=(const Holder& o)
{
	if (this != &o) {
		mFrame = o.mFrame;
		mEntry = o.mEntry;
	}
}
